####################
## This file is part of PBPKxIon package.
##   PBPKxIon is free software: you can redistribute it and/or modify
##   it under the terms of the GNU General Public License as published by
##   the Free Software Foundation, either version 3 of the License, or
##   (at your option) any later version.
##
##   PBPKxIon is distributed in the hope that it will be useful,
##   but WITHOUT ANY WARRANTY; without even the implied warranty of
##   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
##   GNU General Public License for more details.
##
##   You should have received a copy of the GNU General Public License
##   along with PBPKxIon.  If not, see <http://www.gnu.org/licenses/>.
## Author comment: None
## File Description: This file holds the Partition Coefficient class definition, with all associated
##   methods to calculate the PCs via Peyret, Poulin, and Kannan (2010)
#################### 
PartitionCoefficient <- setRefClass("PartitionCoefficient",
  fields = list(
  	P.aplw = "numeric",
  	P.prwp = "numeric",
  	P.prwi = "numeric",
  	P.prwc = "numeric",
  	P.prwe = "numeric",
  	P.ct = "numeric",
  	P.it = "numeric",
  	P.p = "numeric",
  	P.e = "numeric",
  	P.Hbw = "numeric",
  	P.tb = "numeric",
  	P.b = "numeric",
  	K.pu = "numeric",
  	K.p = "numeric",
  	P.bw = "numeric"
  ),
  methods = list(
  	#####
    ## Run the partition coefficients; this sets up the data, taking the chemical parameters and ionization from outside.
    #####
    Tissue.PC.per.Chemical = function(tissue, chemical) {
 	    .self$predPC(tissue, chemical)
      ## Physiological
      ## in PartitionCoefficient$F
    },
    ##########
	### Data should be properly formatted and in their appropriate data.frames. Run PC predictions.
    ##########
    predPC = function (tissue, chemical)
    {
      #####
      ## Calculate the partition coefficients
      #####
      ## IONIZATION
      I <- list(w = chemical$I.m(7.4),
                it = chemical$I.m(tissue$interstitial.fluid$pH),
                ct = chemical$I.m(tissue$cellular.tissue$pH),
                e = chemical$I.m(tissue$erythrocyte$pH),
                p = chemical$I.m(tissue$plasma$pH)
               )
      #####
      ## HYDRO-PHOBICITY/PHILICITY
      #####
      ## Need to elegantly handle Pow/Pvow/logP
      ## Handled: Set all water PCs upon read-in. Now don't need to worry about missing values.
      f_P.ow <- if (identical(tissue$tissueType, "Adipose")) chemical$P.vow else chemical$P.ow
      #####
      ## VOLATILITY (vapor.pressure, in Pa) -- VOC >= 10 Pa : 10^{-9} Pa < SVOC < 10 Pa
      #####
      #####
      ## ET CETERA
      #####
      ## Some chem terms that are compound terms, rather than single fields in the chemical.
      .self$P.bw <- .self$Pbw(tissue, chemical, I)
      #CL.int <- chem.df$CL.int         ## 20150114: Needed for what? Schmitt??

 	    #####
  	  ## A lot of checks to fill in missing data, add characteristics, etc. 
  	  ## B.P should probably be calculated and kept, but at this time, B.P is the chemical 
 	    #####
      if(is.null(chemical$B.P) | length(chemical$B.P) == 0) {
        if (length(chemical$fub) == 0) {
          chemical$fub <- 0.0005
          message("Warning for Blood.plasma partition coefficient.\n  Chemical ", chemical$ID$Compound, " is missing fraction unbound (fub). Assuming fub == 0.0005.\n")
        }
        ## A QSAR built from data on basic drugs from Rodgers, Tozer, & Rowland (2005), by Yun (2014), where (n=28, R^2=0.40). 
        chemical$B.P <- Buist2012_BP(chemical) 
  	  }

      # Acidic phospholipid to water
      .self$P.aplw <- .self$Paplw(tissue, chemical, I)
      # Protein in plasma to water
      .self$P.prwp <- .self$Pprwp(tissue, chemical, I)
      # Protein in interstitial fluid to water (assumed to be same as plasma
      .self$P.prwi <- .self$Pprwi(tissue, chemical, I)
      # Protein in cellular tissue to water (negligible)
      .self$P.prwc <- .self$Pprwc(tissue, chemical, I)

      .self$P.ct <- .self$Pmwct(tissue, chemical, I)
      .self$P.it <- .self$Pmwit(tissue, chemical, I)
      .self$P.p <- .self$Pmwp(tissue, chemical, I)
      .self$P.e <- .self$Pmwe(tissue, chemical, I)
      .self$P.b <- .self$Pb(tissue, chemical, I)
      .self$K.pu <- .self$Kpu(tissue, chemical, I)
      .self$K.p <- .self$Kp(tissue, chemical, I)
      .self$P.tb <- .self$Ptb(tissue, chemical, I)
      #if (chemical$ID$Compound == "Chlorpyrifos") {
 	    #.self$reportPC(tissue, chemical, I)
      #}
    },
    ######### CODE FOR DEALING WITH HLOCs ###########
    ## Need to integrate this with the above functions so it is thoughtless for users.
    #####
    ## General function for the calculation of the Kp. Takes all of the individual arguments, rather than data frames
    ## containing the arguments. Would like to overload this function with the data.frame acceptor.
    ## Similar procedures would be great for the Peyret method, as well.
    #####
    HLOC.Kp = function(tissue, chemical, I)
    {
	  ## Already calculated the P.vow for each compound (if not entered by user) in FillHydroAffinity (Class_chemical.R)
	  ## f_P.ow_ is the P.ow for this function.
	  if(tissue$tissueType == "Adipose") {
        f_P.ow_ <- chemical$P.vow
      } else {
      	f_P.ow_ <- chemical$P.ow
      }

	  f.nlte <- tissue$F[["nlct"]] + 0.3 * tissue$F[["nplct"]]
	  f.nlpe <- tissue$F[["nlp"]] + 0.3 * tissue$F[["nplp"]]
	  max.Kp <- f.nlte / f.nlpe
	  fu.p <- (1 + I[["p"]]) * tissue$F[["wp"]] + f_P.ow_ * f.nlpe
	  ## P.nlw is the "water" partition coefficient. I.e, mostly P.ow, except for Adipose tissue, where it is P.vow
	  fu.t <- (1 + I[["ct"]]) * tissue$F[["wct"]] + f_P.ow_ * f.nlte
#      HLOC.invalidate.PC(tissue, chemical, I)
	  .self$K.p <- fu.t / fu.p
     #cat("f_P.ow_:", f_P.ow_, "\n")
     #cat("f.nlte:", f.nlte, "\n")
     #cat("f.nlpe:", f.nlpe, "\n")
     #cat("max.Kp:", max.Kp, "\n")
     #cat("fu.p:", fu.p, "\n")
     #cat("fu.t:", fu.t, "\n")
     #cat("tissue:", tissue$tissueType, "  vol:", tissue$tissueVolume, "\n")
     #cat("*****self$K.p: ", .self$K.p, "\n")

	  return(.self$K.p)
    },
    ##########
    ### For the tissue:blood partition, we deal with the neutral lipid equivalents, meaning the fraction of neutral lipids * 0.3 fraction 
    ### neutral phospholipids.
    ##########
    HLOC.Patb = function(tissue, chemical, I)
    {
      F.neat <- tissue$F[["nlct"]] + 0.3 * tissue$F[["nplct"]]
      F.neb <- tissue$F[["nle"]] + 0.3 * tissue$F[["nple"]]
      .self$P.tb <- F.neat / F.neb
      return(.self$P.tb)
    },
    ##########
    ## Reports each chemical PCs calculated
    ##########
    reportPC = function(tissue, chemical, I) {
      HLOC <- .self$is.HLOC(tissue, chemical, I)
      cat("Chemical:", chemical$ID$Compound, "Tissue:", tissue$tissueType, "\n")
      cat("P.tb    ", .self$P.tb, "\n")
      cat("K.pu    ", .self$K.pu, "\n")
      cat("K.p     ", .self$K.p, "\n")
      cat("P.ct    ", paste0(if(HLOC) "x", .self$P.ct), "\n")
      cat("P.it    ", paste0(if(HLOC) "x", .self$P.it), "\n")
      cat("P.p     ", paste0(if(HLOC) "x", .self$P.p), "\n")
      cat("P.e     ", paste0(if(HLOC) "x", .self$P.e), "\n")
      cat("P.prwp  ", paste0(if(HLOC) "x", .self$P.prwp), "\n")
      cat("P.prwi  ", paste0(if(HLOC) "x", .self$P.prwi), "\n")
      cat("P.prwc  ", paste0(if(HLOC) "x", .self$P.prwc), "\n")
      cat("P.prwe  ", paste0(if(HLOC) "x", .self$P.prwe), "\n")
      cat("P.aplw  ", paste0(if(HLOC) "x", .self$P.aplw), "\n")
      cat("P.b     ", paste0(if(HLOC) "x", .self$P.b), "\n")
      cat("P.bw    ", paste0(if(HLOC) "x", .self$P.bw), "\n")
      cat("P.Hbw   ", paste0(if(HLOC) "x", .self$P.Hbw), "\n")
    }
  )
)
