####################
### Object for storing ALL chemicals (phys-chem properties, too?)
### See some methods in PBPKxIon (setup methods, read-ins)
####################
Data <- setRefClass("Data",
  fields = list(
    chemList = "list",   ## Should this be a dataframe? Each chemList has number of ionizable sites?
    pKaTools.df = "data.frame",  ## Main data.frame of this analysis.
    predictor = "character",   ## Each dataset is now a predictor.
    ## Holds nonstandard pcps, such as ACToRUseDB terms and ADMET Predictor IATs
    ## Made this part of the Data object to make the data manageable. Need to be sure that an ID is associated with
    ## the proprietary fields, though, or else it is useless.
    proprietary_fields = "data.frame"
  ),
  methods = list(
    ## Classifies chemicals
    setUseCategories = function(categoryList_file = NULL, chem_data) {
      chem_classes <- classify.Chemicals(category_file = categoryList_file, chem_data)
      chem_classes <- mergeAndDrop(chem_data, chem_classes, merge_by="CASRN")
      return(chem_classes)
    },
    #####
    ## Read in selected files and construct the non-flat data frame. This will be used to create the Data() object
    ## which will then be used to create the flat file and run analyses.
    #####
    Initialize_PBPKxIon = function(pKa_file, pKa_sheet="MARVIN", fingerprint_file=NULL, fingerprint_sheet=1, calc_chemical_category=FALSE) {
      ## Get and set all data for pKaTools.
      if (length(.self$chemList) == 0) {
        stop("Initialize_PBPKxIon: MUST SET CHEMICAL LIST BEFORE GETTING pka predictions.\n")
      }
      if (!is.null(pKa_file)) {
       pKa <- setPredictions(pKa_file, pKa_sheet)
       .self$pKaTools.df <- mergeAndDrop(.self$pKaTools.df, pKa)
      }
      if (!is.null(fingerprint_file)) {
        cat("Reading fingerprints:\n")
        fp <- setFingerprints(fingerprint_file, fingerprint_sheet)
        .self$pKaTools.df <- mergeAndDrop(.self$pKaTools.df, fp) ## Temporarily make this a global data.frame so that I can work on it.
      }

      ## Again, if there is a pKa file to predict, then this needs to be set.
      if (!is.null(pKa_file)) {
        ## Fix up the pKa's to be PKATOOLS format.
        if (!is.pKaToolsFormat(.self$pKaTools.df)) {
          .self$pKaTools.df <- normalize_pKa(.self$pKaTools.df)
        }
      }

      ## SPARC comma's placed wherever. Manually removing them and fixing pKax (x=1..6) with multiple values reported (pushing them to
      ## pKa{x+1} and pKa{x+2}, etc, for as many as are listed). 
      #check.pKaTools <<- .self$pKaTools.df
      #cat ("Initializing pKaTools object. Check to see what the first couple of chemicals look like. (check.pKaTools)\n")
      #browser()
      # Create the object!
      .self$pKaTools_Object()

      ####
      ## Below: set the source data, then further classify the chemicals by their source (ToxCast/Tox21)
      ####
      # Set object membership (chemList$Source). Have 2 RData files, ToxCast and Tox21. ToxCast strict subset of Tox21.
      ###
      for (i in 1:length(.self$chemList)) .self$chemList[[i]]$setSource()
      ###
      # Set each chemical to their category
      ###
      if (calc_chemical_category) {
        cats <- setUseCategories(chem_data=.self$pKaTools.df)
        .self$pKaTools.df <- mergeAndDrop(.self$pKaTools.df, cats)
      }
    },
    #####
    ## Create the pKaTools Data object using the data.frame. Incorporate the data.frame into the object for further analyses,
    ## such as flattening the data. (getFlatFile?)
    #####
    pKaTools_Object = function() {
      options(warn=2)   # 1: warnings as they occur, 2: treat warnings as errors
      ## Set up the data set:
      cat("Setting object:\n")
      pb <- txtProgressBar(min=1, max=nrow(.self$pKaTools.df), initial=1, style=3)
      for (j in 1:nrow(.self$pKaTools.df)) {
        setTxtProgressBar(pb, j)
        inpKaTools.j <- which(pKaTools.df$CASRN == .self$chemList[[j]]$ID$CASRN)
        #cat("******************************", j, .self$pKaTools.df[inpKaTools.j,1], .self$pKaTools.df[inpKaTools.j,2], "*************************************\n")
        ## Chemical data, and
        if (j > length(.self$chemList)) .self$setChem(.self$pKaTools.df[j,])
        ## pKa data.
        #cat("** Final pKaTools_Object\n")
        #cat("** ")
        #print(.self$pKaTools.df[inpKaTools.j, c("pKaTools_APKA", "pKaTools_APKA.atoms", "pKaTools_BPKA", "pKaTools_BPKA.atoms")])
        #cat("**")
        #print(.self$pKaTools.df[inpKaTools.j, "SMILES"])
        #cat("**")
        #print(.self$chemList[[j]]$structure$SMILES)
        #cat("**************************************************************************************************************************\n")
        .self$chemList[[j]]$populatePKA(.self$pKaTools.df[inpKaTools.j,,drop=FALSE])
      }
      cat("\n")
    },
    #####
    ## Insert all chemicals from a data frame into the chemList. This is a generally useful function for both reading
    ## in the xls data, as well as creating a new data object from a pKaTools.df
    #####
    SetChemList = function(df)
    {
      cat("Inserting chemical data into data set.\n")
      pb <- txtProgressBar(min=1, max=nrow(df), initial=1, style=3)
      for (i in 1:nrow(df)) {
        setTxtProgressBar(pb, i)
        ## Make sure that if this is a new setting of chemical data, that a container is created for it.
        if (i > length(.self$chemList)) .self$setChem(df[i,])
        #.self$chemList[[i]]$populateChemicalPropertyFields(df[i,])
      }
      cat("\n")
    },
    #####
    ## Setting the predictions in the dataSet object.
    #####
    setPredictions = function(predfile, sheet="MARVIN", indf = NULL) {  ## Assumption of MARVIN to begin.
      ## Gather the predictions/chemical data
      ## If pred exists in .GlobalEnv, and the current value of the global variable \in {SPARC, MARVIN, ADMETPRED}
      ## is set to the same value as the input variable `sheet' above, then this read-in can be skipped. Ideally, 
      ## almost all read-ins can be skipped after the initial read.
      cat("Reading", sheet, "predictions (", predfile, "):\n")
      pred <- read.file(predfile, sheet, sep = "|")
      ## Fix the CAS, as they are the necessary key for all lookups
      CAS.clean <- get.CAS.list(pred$CASRN)
      ## Change CAS in predictions to fixed CAS.
      pred$CASRN <- CAS.clean
      ## Remove duplicates
      pred <- pred[!duplicated(pred, by = "CASRN"),]
      ## Order by CASRN
#      setorder(pred, CASRN)   # data.table parlance for ordering; order 'pred' by 'CASRN'. Add more variables of pred if interested.
      ## Fix up the pKa's to be PKATOOLS format.
      ## Assumption is that the user can input a data.frame with previously read data, and can be merged.
      if (!is.null(indf)) pred <- mergeAndDrop(pred, indf, merge_by="CASRN")
      return(pred)
    },
    #####
    ## Imports the chemical data input for each chemical. Reads the input data, then passes the data to each chemical.
    #####
    ImportChemData = function(filename, sheet="active") {
      cat("Reading in chemical data from", filename, "sheet =", sheet, "\n")
      #chem.data <<- read.xls(filename, sheet=sheet, stringsAsFactors = FALSE)
      chem.data <- read.file(filename, sheet=sheet, sep = "|")

      setThesaurus()
      ## Fix up the pKa's to be PKATOOLS format.
      if (any(c(grepl("PKA", names(chem.data), ignore.case=TRUE), grepl("PKB", names(chem.data), ignore.case=TRUE)))) {
        message("Ignoring input pKa in the chemical data sheet. If these are the desired pKa, place these predictions in a new sheet and\ninput under the label pred.or.input.pKa = <xls_file>, pred.or.input.pKa.sheet = <sheetname>\n")
      }
      ## Next 2 lines: data.table doesn't like names(DT) <- "xxx"; setnames(x,old,new) used instead.
      standard.names <- standardize_chemData(names(chem.data))
      setnames(chem.data, names(chem.data), standard.names)
      ## Charge can be listed in postive/negative/neutral terms, which are non-numeric. This function will make them numeric.
      if ("charge" %in% names(chem.data)) {
        chem.data$charge <- DigitizeCharge(chem.data$charge)
      }
      #####
      ## This can probably be put into a function, since it is the same as setPredictions above.
      #####
      CAS.clean <- get.CAS.list(chem.data$CASRN)
      chem.data$CASRN <- CAS.clean
      chem.data <- chem.data[!duplicated(chem.data, by="CASRN"),]
      setorder(chem.data, CASRN)
      .self$pKaTools.df <- chem.data
      chem.data <<- chem.data
      .self$SetChemList(chem.data)
    },
    #####
    ## setChem: Create a new chemicalProperties object, and initialize the chemProp fields with the incoming df.
    ## Note: To copy chemical, do: chemicalProperties$new(which.chem))
    #####
    setChem = function(chemRow=data.frame) {
      # Create new chemicalProperties object, append it to end of Data field, "chemList"
      .self$chemList <- c(.self$chemList, chemicalProperties$new())
      # Initialize the values of the chemicalProperties object usind the chemProp method "setFields"
      .self$chemList[[length(.self$chemList)]]$populateChemicalPropertyFields(chemRow)
      ## Only set chemical properties from ImportChemData, methinks. Simpler to debug if only one instance, and also comes from the
      ## chem.data sheet of the .xlsx file, rather than the various predictor pages (which will cause the predictions to be incomparable)
    },
    #####
    ## Add individual chemicals to dataset
    #####
    addChem = function(Compound=NULL, CASRN=NULL, GSID=NULL, SMILES=NULL, InChI=NULL, InChIKey=NULL, pKa=NULL, logD=numeric(0), 
                       logPow=numeric(0), logPvow=numeric(0), MW=numeric(0), Henrys_Law=numeric(0), boiling_point=numeric(0), 
                       vapor_pressure=numeric(0), charge=numeric(0), Source=character(0)) 
    {
      chemRow <- list(
        Compound = Compound,
        CASRN = CASRN,
        GSID = GSID,
        SMILES = SMILES,
        InChI = InChI,
        InChIKey = InChIKey,
        logD = logD,
        logPow = logPow,
        logPvow = logPvow,
        MW = MW,
        Henrys_Law = Henrys_Law,
        boiling_point = boiling_point,
        vapor_pressure = vapor_pressure,
        charge = charge,
        Source = Source
      )
      ## Turn all non-null, non <class>(0) parameters into a data.frame to pass to setChem.
      chemRow <- as.data.frame(chemRow[sapply(chemRow,length)>0], stringsAsFactors=FALSE)
      chemRow <- data.frame(chemRow, stringsAsFactors=FALSE)
      .self$setChem(chemRow)
    },
    rmChem = function(identifier) {
      if(!is.null(identifier)) {
        rm_element <- chemIndex(identifier)
        idx <- which(.self$pKaTools.flat$CASRN == .self$chemList[[rm_element]]$ID$CASRN)
        .self$pKaTools.flat <- .self$pKaTools[-idx,]
        .self$chemList[[rm_element]] <- NULL
      } else { cat("Identifier is not in chemical list 'CASRN', 'GSID', or 'Compound'. No element removed.\n") }
    },
    ## INPUT: An identifier, either CASRN, GSID, or Compound
    ## OUTPUT: The index of the chemical in the dataSet.
    chemIndex = function(identifier) {
      ## If there are 2-6 digits -- 2 digits -- 1 digit, a CAS has been passed.
      if (grepl("^\\d{2,6}-\\d\\d-\\d$", identifier)) {
        ## CASRN
        cat("Searching by CASRN.\n")
        idx <- which(lapply(.self$chemList, function(x) x$ID$CASRN == identifier)== TRUE)
      } else if (grepl("^\\d\\d\\d\\d\\d+$", identifier) & identifier > 20000) {
        ## DSSTox_GSID
        cat("Searching by DSSTox_GSID.\n")
        idx <- which(lapply(.self$chemList, function(x) x$ID$GSID == identifier)== TRUE)
      } else {
        ## Chemical name... This is the least reliable, maybe let user know to use at own risk.
        cat("Searching by chemical name.\n")
        idx <- which(lapply(.self$chemList, function(x) x$ID$Compound == identifier)== TRUE)
      }
      if(is.null(idx)) stop("Could not find chemical associated with the identifier", identifier, ".\n")
      return(idx)
    },
    #####
    ## These methods deal with reading in data, in particular, CPCat stuff.
    #####
    ## Simple, probably unnecessary utitlity function.  
    deformat_pKaTools.list = function(pKaToolslist) {
      return(unlist(strsplit(pKaToolslist, ";")))
    },
    #####
    ## The following functions works through the beast of CPCat and sets chemical categories. 
    ## Included in this list of methods since they are called from SetUseCategories (in turn, initialize PBPKxIon)
    #####
    classify.Chemicals = function(category_file="~/Projects/pK-2014-09-15/pK/projects/pKa-format/data/CPCat_dist_2014-05-21/cpcat_chemicals.xlsx", chem_data = NULL) {
      if (!exists("cpcat")) {
          cat ("classify.Chemicals: Reading cpcat...\n")
          #data( envir=environment())
          cpcat <<- read.file(category_file, stringsAsFactors=FALSE, header=TRUE)
          cpcat$CASRN <- gsub("\\s", "", cpcat$CASRN)
        }
      if (is.null(chem_data)) stop("classify.Chemicals requires chem_data.\nData file read in of chem_data does not ensure the appropriate steps have been taken to clean the input data.\n")

      # if (exists("chem_class_presort")) {user.input <- ask("chem_class_presort already calculated. Proceed with current values? [Y|n] "); if (toupper(user.input) %in% c("Y", "", "YES", "YE")); return(chem_class_presort) }
      ## For each chemical, split out all of the CAS #s, check if it is in cpcat, add to data.frame, which will be cbinded at end.
      cpcat_terms <- data.frame(chem_class = rep("", nrow(chem_data)), cpcat_term = rep("", nrow(chem_data)), 
                                is.Rx = rep(FALSE, nrow(chem_data)), is.near = rep(FALSE, nrow(chem_data)), 
                                is.far = rep(FALSE, nrow(chem_data)), stringsAsFactors=FALSE)

      ## Iterate over all chemicals, and get the appropriate data from cpcat.
      grabbed <- vector()
      iter <- nrow(chem_data)
      cat("Setting chemical classes:\n")
      pb <- txtProgressBar(min=1, max=iter, initial=1, style=3)
      for (i in 1:iter) {
        setTxtProgressBar(pb, i)
        grabbed <- NULL
        ## pKaTools standard way of representing a list of anything: a semicolon delimited list.
        x <- deformat_pKaTools.list(as.character(chem_data$CASRN[i]))
        ## get all indices that are associated with the CASRN in the list
        grabbed <- c(grabbed, which(cpcat$CASRN %in% x))
        ## Keep only unique
        capture_terms <- unique(cpcat$Description_CPCAT[grabbed])
        ## Classify the chemical class.
        a <- classifyChem(capture_terms)
        ## Copy the chemical class data.
        cpcat_terms$chem_class[i] <- a$chem_class
        cpcat_terms$cpcat_term[i] <- a$cpcat_term 
        cpcat_terms$is.Rx[i] <- a$is.Rx
        cpcat_terms$is.near[i] <- a$is.near
        cpcat_terms$is.far[i] <- a$is.far
        .self$chemList[[i]]$chem_class <- a$chem_class
        .self$chemList[[i]]$cpcat_term <- a$cpcat_term
        .self$chemList[[i]]$is.Rx <- a$is.Rx
        .self$chemList[[i]]$is.near <- a$is.near
        .self$chemList[[i]]$is.far <- a$is.far
      }
      cat("\n")

      cat("nrow(cpcat_terms: ", nrow(cpcat_terms), ", nrow chem_data: ", nrow(chem_data), "\n")

      ## Major data structure: This contains both the chemical data along with it's cpcat chemical classifications.
      chem_class_presort <- cbind(chem_data, cpcat_terms)
      ## Summary of all chemical classes, venn diagram, etc.
      chem_class_presort <<- chem_class_presort[order(chem_class_presort$CASRN),]
      Summarize_Chemical_Data(chem_class_presort)

      return(chem_class_presort)
    },
    #####
    ## Takes a set of terms for a chemical and sets it to the correct chemical class(es)
    #####
    classifyChem = function(terms) {
      x <- data.frame(chem_class=NA, cpcat_term=NA, is.Rx=FALSE, is.near=FALSE, is.far=FALSE, stringsAsFactors=FALSE)
      is.Rx <- FALSE
      is.near <- FALSE
      is.far <- FALSE
      UseDB_terms <- NULL

      if ("drug_ACToRUseDB" %in% terms) { 
        UseDB_terms <- "drug_ACToRUseDB"
        is.Rx <- TRUE
      }
      if ("fragrance_ACToRUseDB" %in% terms) {
        list_term <- if(!is.null(UseDB_terms)) "," else NULL
        UseDB_terms <- paste0(UseDB_terms, list_term, "fragrance_ACToRUseDB")
        is.near <- TRUE
      } 
      if ("consumer_use_ACToRUseDB" %in% terms) {
        list_term <- if(!is.null(UseDB_terms)) "," else NULL
        UseDB_terms <- paste0(UseDB_terms, list_term, "consumer_use_ACToRUseDB")
        is.near <- TRUE
      } 
      if ("personal_care_ACToRUseDB" %in% terms) {
        list_term <- if(!is.null(UseDB_terms)) "," else NULL
        UseDB_terms <- paste0(UseDB_terms, list_term, "personal_care_ACToRUseDB")
        is.near <- TRUE
      }  
      if ("food_additive_ACToRUseDB" %in% terms) {
        list_term <- if(!is.null(UseDB_terms)) "," else NULL
        UseDB_terms <- paste0(UseDB_terms, list_term, "food_additive_ACToRUseDB")
        is.near <- TRUE
      } 
      if ("antimicrobial_ACToRUseDB" %in% terms) { 
        list_term <- if(!is.null(UseDB_terms)) "," else NULL
        UseDB_terms <- paste0(UseDB_terms, list_term, "antimicrobial_ACToRUseDB") 
        is.far <- TRUE
      } 
      if ("colorant_ACToRUseDB" %in% terms) { 
        list_term <- if(!is.null(UseDB_terms)) "," else NULL
        UseDB_terms <- paste0(UseDB_terms, list_term, "colorant_ACToRUseDB") 
        is.far <- TRUE
      } 
      if ("flame_retardant_ACToRUseDB" %in% terms) { 
        list_term <- if(!is.null(UseDB_terms)) "," else NULL
        UseDB_terms <- paste0(UseDB_terms, list_term, "flame_retardant_ACToRUseDB") 
        is.far <- TRUE
      } 
      if ("industrial_manufacturing_ACToRUseDB" %in% terms) { 
        list_term <- if(!is.null(UseDB_terms)) "," else NULL
        UseDB_terms <- paste0(UseDB_terms, list_term, "industrial_manufacturing_ACToRUseDB") 
        is.far <- TRUE
      } 
      if ("petrochemical_ACToRUseDB" %in% terms) { 
        list_term <- if(!is.null(UseDB_terms)) "," else NULL
        UseDB_terms <- paste0(UseDB_terms, list_term, "petrochemical_ACToRUseDB") 
        is.far <- TRUE
      } 
      if ("chemical_warfare_ACToRUseDB" %in% terms) { 
        list_term <- if(!is.null(UseDB_terms)) "," else NULL
        UseDB_terms <- paste0(UseDB_terms, list_term, "chemical_warfare_ACToRUseDB") 
        is.far <- TRUE
      } 
      if ("fertilizer_ACToRUseDB" %in% terms) { 
        list_term <- if(!is.null(UseDB_terms)) "," else NULL
        UseDB_terms <- paste0(UseDB_terms, list_term, "fertilizer_ACToRUseDB") 
        is.far <- TRUE
      } 
      if ("herbicide_ACToRUseDB" %in% terms) { 
        list_term <- if(!is.null(UseDB_terms)) "," else NULL
        UseDB_terms <- paste0(UseDB_terms, list_term, "herbicide_ACToRUseDB") 
        is.far <- TRUE
      } 
      if ("inert_ACToRUseDB" %in% terms) { 
        list_term <- if(!is.null(UseDB_terms)) "," else NULL
        UseDB_terms <- paste0(UseDB_terms, list_term, "inert_ACToRUseDB") 
        is.far <- TRUE
      } 
      if ("pesticide_ACToRUseDB" %in% terms) { 
        list_term <- if(!is.null(UseDB_terms)) "," else NULL
        UseDB_terms <- paste0(UseDB_terms, list_term, "pesticide_ACToRUseDB") 
        is.far <- TRUE
      } 
      if (is.null(UseDB_terms)) { 
        UseDB_terms <- "NONE" 
        is.far <- TRUE
      }

      if (is.Rx) { x$chem_class <- "Rx" }
      else if (is.near) { x$chem_class <- "env.near" }
      else { x$chem_class <- "env.far" }
      x$cpcat_term <- UseDB_terms
      x$is.Rx <- is.Rx
      x$is.near <- is.near
      x$is.far <- is.far
      #cat(x$cpcat_term, "Rx:", x$is.Rx, "near:", x$is.near, "far:", x$is.far, "\n")

      return(x)
    },
    #####
    ## Make table of chemical classes by dataset.
    #####
    Summarize_Chemical_Data = function(chem_class_data = NULL) {
      if (is.null(chem_class_data)) {
        if (!is.null(chem_class_presort))
          chem_class_data <- chem_class_presort
        else {
          err <- paste("Summarize_Chemical_Data requires either:\n",
                       "* chem_class_presort in the .GlobalEnv, or:\n",
                       "* An argument with a data frame that holds chemical data with the following information:\n",
                       "  - cpcat_term: The ACToRUseDB term associated with chemical\n",
                       "  - is.Rx, is.near, is.far: Logical terms for the classification of the chemical into pharma, near, and far-field, respectively\n",
                       "Calculting necessary information:\n"
                      )
          cat(err) 
          .self$classify.Chemicals()
        }
      }

      ###
      # Grab the categories of ToxCast and Tox21 chemicals
      ###
      # True/False vector, saying whether chemical is a ToxCast chemical or not.
      ToxCast.tf <- lapply(.self$chemList, FUN = function(x) { if(x$Source == "ToxCast") TRUE else FALSE })
      Tox21.tf <- lapply(.self$chemList, FUN = function(x) { if(x$Source == "ToxCast" | x$Source == "Tox21") TRUE else FALSE })

      ToxCast.chem <- chem_class_data[unlist(ToxCast.tf),]
      Tox21.chem <- chem_class_data[unlist(Tox21.tf),]

      useDB.table <- data.frame(
                    Use.Category = c("antimicrobial_ACToRUseDB", "colorant_ACToRUseDB", "drug_ACToRUseDB", 
                             "flame_retardant_ACToRUseDB", "fragrance_ACToRUseDB","industrial_manufacturing_ACToRUseDB", 
                             "personal_care_ACToRUseDB", "petrochemical_ACToRUseDB", "chemical_warfare_ACToRUseDB", 
                             "consumer_use_ACToRUseDB", "fertilizer_ACToRUseDB", "food_additive_ACToRUseDB", 
                             "herbicide_ACToRUseDB", "inert_ACToRUseDB", "pesticide_ACToRUseDB", "NO USE CATEGORY"
                            ),
                    Assignment = c("Far-field", "Far-field", "Pharmaceutical",
                             "Far-field", "Near-field", "Far-field",
                             "Near-field", "Far-field", "Far-field",
                             "Near-field", "Far-field", "Near-field",
                             "Far-field", "Far-field", "Far-field", "Far-field"
                            ),
                    Counts_Exposure_Universe = c(
                           length(which(grepl("antimicrobial_ACToRUseDB", chem_class_data$cpcat_term))),
                           length(which(grepl("colorant_ACToRUseDB", chem_class_data$cpcat_term))),
                           length(which(grepl("drug_ACToRUseDB", chem_class_data$cpcat_term))),
                           length(which(grepl("flame_retardant_ACToRUseDB", chem_class_data$cpcat_term))),
                           length(which(grepl("fragrance_ACToRUseDB", chem_class_data$cpcat_term))),
                           length(which(grepl("industrial_manufacturing_ACToRUseDB", chem_class_data$cpcat_term))),
                           length(which(grepl("personal_care_ACToRUseDB", chem_class_data$cpcat_term))),
                           length(which(grepl("petrochemical_ACToRUseDB", chem_class_data$cpcat_term))),
                           length(which(grepl("chemical_warfare_ACToRUseDB", chem_class_data$cpcat_term))),
                           length(which(grepl("consumer_use_ACToRUseDB", chem_class_data$cpcat_term))),
                           length(which(grepl("fertilizer_ACToRUseDB", chem_class_data$cpcat_term))),
                           length(which(grepl("food_additive_ACToRUseDB", chem_class_data$cpcat_term))),
                           length(which(grepl("herbicide_ACToRUseDB", chem_class_data$cpcat_term))),
                           length(which(grepl("inert_ACToRUseDB", chem_class_data$cpcat_term))),
                           length(which(grepl("pesticide_ACToRUseDB", chem_class_data$cpcat_term))),
                           length(which(grepl("NONE", chem_class_data$cpcat_term)))
                          ),
                    Counts_ToxCast = c(
                           length(which(grepl("antimicrobial_ACToRUseDB", ToxCast.chem$cpcat_term))),
                           length(which(grepl("colorant_ACToRUseDB", ToxCast.chem$cpcat_term))),
                           length(which(grepl("drug_ACToRUseDB", ToxCast.chem$cpcat_term))),
                           length(which(grepl("flame_retardant_ACToRUseDB", ToxCast.chem$cpcat_term))),
                           length(which(grepl("fragrance_ACToRUseDB", ToxCast.chem$cpcat_term))),
                           length(which(grepl("industrial_manufacturing_ACToRUseDB", ToxCast.chem$cpcat_term))),
                           length(which(grepl("personal_care_ACToRUseDB", ToxCast.chem$cpcat_term))),
                           length(which(grepl("petrochemical_ACToRUseDB", ToxCast.chem$cpcat_term))),
                           length(which(grepl("chemical_warfare_ACToRUseDB", ToxCast.chem$cpcat_term))),
                           length(which(grepl("consumer_use_ACToRUseDB", ToxCast.chem$cpcat_term))),
                           length(which(grepl("fertilizer_ACToRUseDB", ToxCast.chem$cpcat_term))),
                           length(which(grepl("food_additive_ACToRUseDB", ToxCast.chem$cpcat_term))),
                           length(which(grepl("herbicide_ACToRUseDB", ToxCast.chem$cpcat_term))),
                           length(which(grepl("inert_ACToRUseDB", ToxCast.chem$cpcat_term))),
                           length(which(grepl("pesticide_ACToRUseDB", ToxCast.chem$cpcat_term))),
                           length(which(grepl("NONE", ToxCast.chem$cpcat_term)))
                          ),
                    Counts_Tox21 = c(
                           length(which(grepl("antimicrobial_ACToRUseDB", Tox21.chem$cpcat_term))),
                           length(which(grepl("colorant_ACToRUseDB", Tox21.chem$cpcat_term))),
                           length(which(grepl("drug_ACToRUseDB", Tox21.chem$cpcat_term))),
                           length(which(grepl("flame_retardant_ACToRUseDB", Tox21.chem$cpcat_term))),
                           length(which(grepl("fragrance_ACToRUseDB", Tox21.chem$cpcat_term))),
                           length(which(grepl("industrial_manufacturing_ACToRUseDB", Tox21.chem$cpcat_term))),
                           length(which(grepl("personal_care_ACToRUseDB", Tox21.chem$cpcat_term))),
                           length(which(grepl("petrochemical_ACToRUseDB", Tox21.chem$cpcat_term))),
                           length(which(grepl("chemical_warfare_ACToRUseDB", Tox21.chem$cpcat_term))),
                           length(which(grepl("consumer_use_ACToRUseDB", Tox21.chem$cpcat_term))),
                           length(which(grepl("fertilizer_ACToRUseDB", Tox21.chem$cpcat_term))),
                           length(which(grepl("food_additive_ACToRUseDB", Tox21.chem$cpcat_term))),
                           length(which(grepl("herbicide_ACToRUseDB", Tox21.chem$cpcat_term))),
                           length(which(grepl("inert_ACToRUseDB", Tox21.chem$cpcat_term))),
                           length(which(grepl("pesticide_ACToRUseDB", Tox21.chem$cpcat_term))),
                           length(which(grepl("NONE", Tox21.chem$cpcat_term)))
                          )
                   )

      print(useDB.table)
      totalChemAndVenn(chem_class_data, "ExpU")
      totalChemAndVenn(ToxCast.chem, "ToxCast")
      totalChemAndVenn(Tox21.chem, "Tox21")

    }
  ) 
)

##########
## Additional Methods. May or may not be useful, but good to do anyways.
##########
Data$methods(
  #####
  ## Calculate the number of ionizable sites for a given data set
  ## This is used to pre-declare the number of rows we will need in a data.frame, e.g., a flat file that has one
  ## row per pKa.
  #####
  getIonizableSitesN = function(countNULL=FALSE) {
    ionizable.atoms <- 0
    for (i in 1:length(chemList)) {
      new_pKa <- chemList[[i]]$getNumIonizableSites()
      ## For flat files, we include all chemicals, even those without pKa. 
      if (countNULL & new_pKa == 0) new_pKa <- 1
      ionizable.atoms <- ionizable.atoms + new_pKa
    }
    return(ionizable.atoms)
  },
  #####
  ## Create data frame with the necessary pKa/SMILES/position/type information. Useful for inferring what atom type
  ## a particular position is (maybe incorporate an atom type routine?? A lot of work, but hella useful.
  ## Data frame of the information, with 1 pKa per line.
  #####
  flatten.pKa = function() {
    total_pKa <- .self$getIonizableSitesN(countNULL=TRUE)
    #cat("************************************************total_pKa: ", total_pKa, "*********************************************\n")
    df <- data.frame(CASRN=rep("", total_pKa), 
                     pKa=rep(NA,total_pKa), 
                     SMILES=rep("",total_pKa), 
                     SMILES_position=rep(0,total_pKa), 
                     class=rep(INFER,total_pKa),
                     stringsAsFactors=FALSE
                    )
 
    ## Counter of which record we are on. At the end of this loop, it should be the same as total_pKa.
    record.no <- 1
    pb <- txtProgressBar(min=1, max=length(chemList), initial=1, style=3)
    for (i in 1:length(chemList)) {
      setTxtProgressBar(pb, i)
      N.pKa <- chemList[[i]]$getNumIonizableSites()
      #cat(2, "i:", i, "has", N.pKa, "pKas\n")
      ## If there are no pKa's, that does not mean we do not want it in the flat file. Just want pKa to be NULL
      if (N.pKa == 0) {
        #cat(record.no, "/", total_pKa)
        df$pKa[record.no] <- NA
        df$SMILES[record.no] <- chemList[[i]]$structure$getSMILES()
        df$SMILES_position[record.no] <- NA
        df$class[record.no] <- NA
        df$CASRN[record.no] <- chemList[[i]]$ID$getCASRN()
        record.no <- record.no + 1
      } else {
        for (j in 1:N.pKa) {
          #cat(record.no, "(", j, ")", "/", total_pKa)
          x <- chemList[[i]]$getChemPKA(j)
          df$pKa[record.no] <- x$value
          df$SMILES[record.no] <- x$SMILES
          df$SMILES_position[record.no] <- x$SMILES_position
          df$class[record.no] <- x$class
          df$CASRN[record.no] <- chemList[[i]]$ID$getCASRN()
          record.no <- record.no + 1
        }
      }
    }
    cat("\n")
    pKaTools.dt <- data.table(.self$pKaTools.df)
    ## Get tons of data all collated. This will add all of the fp, cats, and other chem data.
    df <- mergeAndDrop(df, pKaTools.dt, merge_by="CASRN")
    return(df)
  },
  flatten.CASRN = function() {
    library(dplyr)
    library(tidyr)
    #####
    ## tidyr version of flattening by CASRN ##
    ## Transform CASRN column into a element vector, then for each element, make a copy of the row and place one of the elements
    ## as the CASRN; i.e., flatten the CASRN column.
    #####
    ret.df <- .self$pKaTools.df %>% transform(CASRN = strsplit(CASRN, ";")) %>% unnest(CASRN)
  }

)

