##########
### Class definition for chemical ionization information
###
### To test results in RStudio, each time, run:
###   > source("pKa.R")
###   > x <- ionization(i__pKa__ = <data>, ...)
###   > x$plot_species()
##########
pKa <- setRefClass("pKa",     ## Class name
  fields = list(
    # "public/private"
    ## Data frame for storing pKa's.
    ## Fields:
    ## * value = Numeric pKa value
    ## * type = Acidic pKa or basic pKa. ATOM SPECIFIC, NOT CHEMICAL. Dissociative (goes negative) or associative (goes positive), resp.
    ## * SMILES_position = Atom in the SMILES string that is ionized. Set to '0' if atom is not specified.
    ## * ionizedSMILES = Ionized SMILES string for this particular pKa. If SMILES_position = '0', this is the input SMILES string.
    pKa_df="data.frame",
    pKa_SMILES = "SMILES",
    predictor="character"
  ),
  methods=list(
    #####
    ## A utility function that returns all pKa values (only numbers). Good for plotting pKa's.
    #####
    return.pKa = function() {
      p <- vector()
      if (nrow(pKa_df) > 0) {
        for (i in 1:nrow(pKa_df)) {
          p <- c(p, pKa_df$value)
        }
      }
      return(p)
    },
    #####
    ## Calculates the fraction of chemical species that have a specific charge state (for a set of pKa).
    ## Neutral determined by?
    #####
    frac_species = function(pH, which.species, pKa) {
      build.frac <- 1
      for(i in 1:length(pKa)) {
        sum.pKa <- 0
        for (j in 1:i) {
          sum.pKa <- sum.pKa + pKa[j]
        }
        build.frac <- build.frac + 10^(i*pH-sum.pKa)
        if (which.species == 0) numerator = 1
        if (i == which.species) numerator <- 10^(i*pH-sum.pKa)
      }
      return(numerator/build.frac)
    }
  )
)

#########
# Plotting
#########
pKa$methods(
  ## Plot the fraction of each microspecies on the pH scale, i.e., chemicalize ##
  plot_species = function() {
    color.scheme <- c("red", "black", "green", "purple", "chartreuse4", "azure4", "orange", "indianred3", "hotpink4", "maroon", "steelblue3")

    i__pKa__ <- as.vector(pKa_df$value)
    i__pKa__ <- as.numeric(i__pKa__)
    i__pKa__ <- sort(i__pKa__)

    #cat("before ggplotting\n")

    n <- length(i__pKa__)
    df <- data.frame(x=c(0,14))
    if(n == 0) return(NA)
    if(n > 0) p <- ggplot(df, aes(x)) + stat_function(fun=function(x) frac_species(x, 0, i__pKa__), color = "blue")
    if(n >= 1) p <- p + stat_function(fun=function(x) frac_species(x, 1, i__pKa__), color=color.scheme[1])
    if(n >= 2) p <- p + stat_function(fun=function(x) frac_species(x, 2, i__pKa__), color=color.scheme[2])
    if(n >= 3) p <- p + stat_function(fun=function(x) frac_species(x, 3, i__pKa__), color=color.scheme[3])
    if(n >= 4) p <- p + stat_function(fun=function(x) frac_species(x, 4, i__pKa__), color=color.scheme[4])
    if(n >= 5) p <- p + stat_function(fun=function(x) frac_species(x, 5, i__pKa__), color=color.scheme[5])
    if(n >= 6) p <- p + stat_function(fun=function(x) frac_species(x, 6, i__pKa__), color=color.scheme[6])
    if(n >= 7) p <- p + stat_function(fun=function(x) frac_species(x, 7, i__pKa__), color=color.scheme[7])
    if(n >= 8) p <- p + stat_function(fun=function(x) frac_species(x, 8, i__pKa__), color=color.scheme[8])
    if(n >= 9) p <- p + stat_function(fun=function(x) frac_species(x, 9, i__pKa__), color=color.scheme[9])
    if(n >= 10) p <- p + stat_function(fun=function(x) frac_species(x, 10, i__pKa__), color=color.scheme[10])
    if(n >= 11) p <- p + stat_function(fun=function(x) frac_species(x, 11, i__pKa__), color=color.scheme[11])
    if(n >= 12) warning("plotting curve for only first 12 pKa's")
   
    ## Add titles and axis labels:
    p <- p 
         #+ ggtitle(predictor)        ## Need to add this, probably??
         + xlab("pH") 
         + ylab("Microspecies Fraction") 
         + scale_x_continuous(breaks = round(seq(min(df$x), max(df$x), by=1),1))

    return(p)
  }
)

##########
## Determination of pKa data type, and filling of data.frame
##########
pKa$methods(
  ##########
  ## A small routine to determine the class of pKa, acidic or basic
  ## If a position has a tendency to go positive, consider it basic. Tendency towards negative, acidic. 
  ##########
  which.pKa = function(pre = NULL, post = NULL) {
    if (pre == "nn") { if (post == "n") return(BASE) else return(ERROR) }
    else if (pre == "n") if(post == "nn") return(ACID) else if (post == "N") return(ACID) else return(ERROR)
    else if (pre == "N") if(post == "n") return(ACID) else if (post == "p") return(BASE) else return(ERROR)
    else if (pre == "p") if(post == "N") return(BASE) else if (post == "pp") return(BASE) else return(ERROR)
    else if (pre == "pp") if(post == "p") return(BASE) else return(error)
    else if (is.na(pre) && is.na(post)) return(ERROR)   ## X: error
    else return(INFER)         ## I: infer
  },
  ##########
  ### Assign class of pKa to infer/error
  ##########
  assign.pKa.class = function (in_df = data.frame) {
    infer <- which(in_df$class == INFER)
    error <- which(in_df$class == ERROR)
    if (length(error) > 0) {
      cat("The following pKa entries are erroneous, and chemical class will be inferred assuming compound is acidic or a weak base: ")
      print(in_df[error,])
      infer <- c(infer, error)
    }
    if (length(infer) > 0) {
      for (i in 1:length(infer)) {
        if (in_df$value[i] < 7.4) in_df$class[i] <- 0   ## Assumption of weak base
        else in_df$class[i] <- 1
      }
    }
    return(in_df)
  },
  ##########
  ## This function will get the pKa in the format proposed in manuscript, although without the position info:
  # xpKay(pos)
  # x,y in {nn, n, N, p, pp}, where nn = 2-, n = -, N = neutral, p = +, pp = 2+
  ##########
  pKa_deformat = function(formatted.pKa) {
    #cat(formatted.pKa)
    ## NA short circuit.
    if (is.na(formatted.pKa)) {
    df <- data.frame(value=NULL, class=NULL, SMILES.position=NULL, ionizedSMILES=NULL, IAT.type=NULL, stringsAsFactors=FALSE)
    } else {
      x <- strsplit(as.character(formatted.pKa), ',')
      y <- unlist(x)
      #cat("The pKa's entered: ", y, ", ", length(y), "\n")
      df <- data.frame(value=rep(0,length(y)), class=rep(3,length(y)), SMILES.position=rep(0,length(y)), ionizedSMILES = rep("", length(y)), IAT.type=rep("", length(y)), stringsAsFactors=FALSE)
      if(length(y)) {
        for (i in 1:length(y)) {
          x <- strsplit(as.character(y[i]), ':')
          #print(x)
          z <- unlist(x)
          if(length(z) == 1) {    ## pKa only
            #cat ("Value: ", z, " returned for z. Check to see if you have entered a column of pKa's from a data frame (e.g., in calc$
            df$value[i] <- as.numeric(z)
            df$class[i] <- INFER
            df$SMILES.position[i] <- 0
          } else if (length(z) == 3) {    ## charge.before : pKa : charge.after
            df$value[i] <- as.numeric(z[2])
            df$class[i] <- which.pKa(pre=z[1], post=z[3])
          } else if (length(z) == 4) {    ## Full formulation.
            df$value[i] <- as.numeric(z[2])
            df$class[i] <- which.pKa(pre=z[1], post=z[3])
            df$SMILES.position[i] <- as.numeric(z[4])
          } else {
            cat("Unsure how to handle ", z, ". Ignoring.\n", sep = "")
          }
        }
      }
      df <- assign.pKa.class(df)
    }
    return(df)
  },
  #####
  ## INPUT: The column names of the input data frame with normalized PKA data.
  ## This function is called from pKaConstruct as a method to create the data frame for acids and bases, with the primary
  ## focus on correctly placing the ionizations into the SMILES string.
  ## This function deparses the normalized pKa columns, matching the values, types, and positions of the pKas.
  ## OUTPUT: Sets the pKa data frame for chemical.
  #####
  setpKa = function(df=data.frame(), inSMILES=NULL) {
    # Experimental: pKa only.
    # * pKa_only means that it could be in several formats:
    #   - b : pKa : a : SMILES_position             
    #   - b : pKa : a                       * Infer SMILES_position
    #   - pKa                               * Infer SMILES_position + class
    #   pKa_deformat (from previous incarnation of this tool, ionization.R) handles exactly these data.
    #cat("Experimental\n")
    #.self$pKa_df <- pKa_deformat(df$pKa)
    
    ## Set SMILES object:
    .self$pKa_SMILES <- SMILES$new(SMILES=inSMILES)

    #cat("INSMILES:::::", inSMILES, "\n")

    # For all pKa predictors that report their predictions in a specific format, once this section of the code is reached,
    # format should be standardized:
    apKa_values <- NULL
    apKa_class <- NULL
    apKa_atoms <- NULL
    if (df$pKaTools_APKA != "") {
      apKa_values <- as.numeric(unlist(strsplit(as.character(df$pKaTools_APKA), ";")))
      apKa_class <- rep(ACID, length(apKa_values))
      apKa_atoms <- unlist(strsplit(as.character(df$pKaTools_APKA.atoms), ";"))
    }
    bpKa_values <- NULL
    bpKa_class <- NULL
    bpKa_atoms <- NULL
    if (df$pKaTools_BPKA != "") {head
      bpKa_values <- as.numeric(unlist(strsplit(as.character(df$pKaTools_BPKA), ";")))
      bpKa_class <- rep(BASE, length(bpKa_values))
      bpKa_atoms <- unlist(strsplit(as.character(df$pKaTools_BPKA.atoms), ";"))
    }

    ## If there are no pKa's, then we don't set pKas.
    if (df$pKaTools_APKA == "" & df$pKaTools_BPKA == "") {
      return()
    }
    #print(df)
    #cat("setpKa: What is the class of df AND SMILES (inSMILES; .self$pKa_SMILES???\n")
    #cat("ionization test: pKaTools_APKA \"", df$pKaTools_APKA, "\"   pKaTools_BPKA: \"", df$pKaTools_BPKA, "\"\n")
    .self$pKa_df <- pKa_construct(c(apKa_values, bpKa_values), c(apKa_class, bpKa_class), c(apKa_atoms, bpKa_atoms))
  },
  #####
  ## Creates the pKa data frame. Input is used as the data frame elements, but ionized SMILES string is constructed
  ## using further function calls.
  #####
  pKa_construct = function(values, class, atoms) {
    # Adjust SMILES_position for SPARC (and any others). Such a garbage program...
    if (SPARC) atoms <- fixSPARClikePositions(atoms)
    # The data frame to be returned.
    #cat("********************************************", values, "\n")
    df <- data.frame(value = values, class = class, SMILES_position = as.numeric(atoms), 
                     ionizedSMILES = rep("",length(values)), IAT.type = rep("", length(values)),
                     stringsAsFactors=FALSE)
    if (length(values) == 0) { return(df) }

    if (MARVIN) {
      ## SPARC chokes here, ADMET Pred doesn't predict sites, hence hard to ionize SMILES ##
      for (i in 1:nrow(df)) {
        #print(df)
        ## Yet again making sure that non-reported positions do not try to ionize SMILES strings.
        if (df$SMILES_position[i] != 0) {
          df$ionizedSMILES[i] <- .self$pKa_SMILES$ionize(df$SMILES_position[i], df$class[i])
        }
      }
      ## Identify the IAT type for each of the ionized SMILES strings. "X" is none? We'll see...
      for (i in 1:nrow(df)) {
        # cat("ionized smiles to testIonization: ", df$ionizedSMILES[i], "\n")
        df$IAT.type[i] <- testIonization(df$ionizedSMILES[i], paste.it = TRUE)
      }
    }

    return(df)
  },
  #####
  ## It will come as little surprise that the reported positions of the SPARC data is nonsensical. As opposed to a sensible approach
  ## of counting ATOMS in the SMILES string, and reporting the position as the ATOM number, SPARC instead is designed to report the
  ## position as the ALPHABETICAL position, i.e., the atom silicon counts as TWO positions, despite being a single atom. It only took
  ## 2 wasted days to figure out this SNAFU. yay.
  #####
  fixSPARClikePositions = function(pos) {
    splitted_str <- .self$pKa_SMILES$compoundStrSplit()
    adjustment <- rep(0, length(splitted_str))
    adjustment[1] <- length(splitted_str[1])-1
    for(i in 2:length(splitted_str)) adjustment[i] <- adjustment[i-1] + nchar(splitted_str[i])-1
    for(i in 1:length(pos)) {
      ## ADMET Predictor (and ACD) do no report positions, which means that the position is set to 0 in pKaTools. Skip these.
      if (pos[i] != 0) pos[i] <- as.numeric(pos[i]) - adjustment[as.numeric(pos[i])]
    }
    return(pos)
  },
  #####
  ## Following is a set of functions that return all different aspects of the pKa. Could use the eval(parse(...))
  #####
  getPKAValue = function(which.pKa) {
    #cat("\nPKAVALUE>>>>>> which_pKa:", which.pKa, "pKa[", which.pKa, "]:", pKa_df$value[which.pKa], "all pKa:", pKa_df$value, "\n")
    #stmt <-  eval(parse(text=all_plot))
    return(pKa_df$value[which.pKa])
  },
  getPKAClass = function(which.pKa) {
    return(pKa_df$class[which.pKa])
  },
  getPKASMILES = function(which.pKa) {
    return(pKa_df$ionizedSMILES[which.pKa])
  },
  getPKASMILESPosition = function(which.pKa) {
    return(pKa_df$SMILES_position[which.pKa])
  }
)

