####################
##### Partition Coefficient calculation routines.
## All partition coefficients are passed the same data. Because of this, the user can call any of the functions, and this package
## will gather all of the PCs needed to calculate the asked for PC.
####################
PartitionCoefficient$methods(
  ##########
  ## This is the predicted PC.
  ##########
  Ptb = function(tissue, chemical, I)
  {
    ## First check is to see if this is a Highly-Lipophilic Organic Compound.
    ## If so, remaining calculations do not matter.
    if (.self$is.HLOC(tissue, chemical, I)) {
      .self$P.tb <- .self$HLOC.Patb(tissue, chemical, I)
      return(.self$P.tb)
    }

    ## Check to see if all of the necessary partition coefficients are set
    .self$K.pu <- if (length(.self$K.pu) == 0) .self$Kpu(tissue, chemical, I) else .self$K.pu
    .self$P.b <- if (length(.self$P.b) == 0) .self$Pb(tissue, chemical, I) else .self$P.b
    .self$P.tb <- .self$K.pu / .self$P.b
    return(.self$P.tb)
  },
  #####
  ## Plasma unbound partition coefficient
  #####
  Kpu = function(tissue, chemical, I)
  {
    ## First check is to see if this is a Highly-Lipophilic Organic Compound.
    ## If so, remaining calculations do not matter.
    if (.self$is.HLOC(tissue, chemical, I)) {
      .self$K.pu <- .self$HLOC.Kp(tissue, chemical, I)
      return(.self$K.pu)
    }

    .self$P.ct <- if (length(.self$P.ct) == 0) .self$Pmwct(tissue, chemical, I) else .self$P.ct
    .self$P.it <- if (length(.self$P.it) == 0) .self$Pmwit(tissue, chemical, I) else .self$P.it
    .self$K.pu <- tissue$F[["ct"]] * .self$P.ct + tissue$F[["it"]] * .self$P.it

    #cat('.self$K.pu <- tissue$F[["ct"]] * .self$P.ct + tissue$F[["it"]] * .self$P.it\n')
    #cat('K.pu <- ', tissue$F[["ct"]], '*', .self$P.ct, '+', tissue$F[["it"]], '*', .self$P.it, '\n---------------\n')
    return(.self$K.pu)
  },
  #####
  ## tissue:plasma PC
  #####
  Kp = function(tissue, chemical, I)
  {
    .self$K.pu <- if (length(.self$K.pu) == 0) .self$Kpu(tissue, chemical, I) else .self$K.pu
    .self$P.p <- if (length(.self$P.p) == 0) .self$Pp(tissue, chemical, I) else .self$P.p
    if (.self$is.HLOC(tissue, chemical, I)) {
      .self$K.p <- .self$K.pu
    } else {
      .self$K.p <- .self$K.pu / .self$P.p
    }
    return(.self$K.p)
  },
  #####
  ## Blood partition coefficient
  #####
  Pb = function(tissue, chemical, I)
  {
    .self$P.e <- if (length(.self$P.e) == 0) .self$Pmwe(tissue, chemical, I) else .self$P.e
    .self$P.p <- if (length(.self$P.p) == 0) .self$Pmwp(tissue, chemical, I) else .self$P.p
    .self$P.b <- tissue$F[["e"]] * .self$P.e + tissue$F[["p"]] * .self$P.p
    return(.self$P.b)
  },
  ##########
  ## In Peyret's tables, PHbw is simply a set value. It is calculable in Peyret's paper, so do that. For now, just set to Peyret's table values.
  ## "For erythrocyte, Pprw refers to the hemoglobin:water PC (PHbw) only for relatively hydrophobic VOCs"
  ## * What is 'relatively hydrophobic'?
  ## * How is VOC defined (Using John Little's (VaTech) as the working definition)
  ##########
  ## This function produces error. For the most part, there is < 5% error (with a number of chemicals with < 1%). However, for the pentane's and Cyclohexane, 
  ## there is upwards of %50 error.
  ##########
  PHbw = function(tissue, chemical, I)
  {
    neutral.lipid.contribution <- .self$POW(tissue, chemical, I) * tissue$F[["nle"]]
    .self$P.bw <- .self$Pbw(tissue, chemical, I)
    .self$P.aplw <- if (length(.self$P.aplw) == 0) .self$Paplw(tissue, chemical, I) else .self$P.aplw
    .self$P.p <- if (length(.self$P.p) == 0) .self$Pmwp(tissue, chemical, I) else .self$P.p
    ## Complicated to put onto 1 line, splitting it up here.
    numerator.1.1 <- (.self$P.bw - (1.0 - tissue$F[["e"]]) * .self$P.p) / tissue$F[["e"]]
    numerator.1.2 <- 1 + I[["w"]]
    numerator.2.1 <- (1 + I[["e"]]) * tissue$F[["we"]]
    numerator.3.1 <- I[["e"]] * .self$P.aplw * tissue$F[["aple"]]
    denom <- tissue$F[["pre"]]
    numerator <- (numerator.1.1 * numerator.1.2 - (numerator.2.1 + neutral.lipid.contribution + numerator.3.1))
    .self$P.Hbw <- numerator / denom

    #if (chemical$ID$CASRN == "26675-46-7" & tissue$tissueType == "xLiver") {
    #  cat("P.Hbw data:\n")
    #  cat("P.bw:", .self$P.bw, "\n")
    #  cat("P.p:", .self$P.p, "\n")
    #  cat("P.aplw:", .self$P.aplw, "\n")
    #  cat("F.e:", tissue$F[["e"]], "\n")
    #  cat("I.w:", I[["w"]], "\n")
    #  cat("I.e:", I[["e"]], "\n")
    #  cat("Faple:", tissue$F[["aple"]], "\n")
    #  cat("F.pre:", tissue$F[["pre"]], "\n")
    #  #browser()
    #}
    if(.self$P.Hbw < 0) .self$P.Hbw <- 0

    return(.self$P.Hbw)
  },
  ##########
  ## Blood water partition coefficient. 
  ##########
  Pbw = function(tissue, chemical, I) 
  {
    ## P.bw = blood/air * air/water
    ## Air water is KAW, or Henrys Law Constant (unitless)
    if (chemical$KAW == 0) {
      .self$P.bw <- 0
    } else {
      .self$P.bw <- chemical$blood.air * chemical$KAW
    }

    return(.self$P.bw)
  },
  #################### MW for the four matrices ####################
  ##########
  ## The following two functions calculate of partition coefficients for tissue matrices, i.e. {ct, e}, and water matrices, {it, p}.
  ## This is due to water matrices containing albumin and lipoproteins, versus the catch-all protein in tissue.
  ##########
  Pmwct = function(tissue, chemical, I)
  {
    .self$P.aplw <- if (length(.self$P.aplw) == 0) .self$Paplw(tissue, chemical, I) else .self$P.aplw
    .self$P.prwc <- if (length(.self$P.prwc) == 0) .self$Pprwc(tissue, chemical, I) else .self$P.prwc

    # new method...
    P.mw <- (tissue$F[["wct"]] + 0.7 * tissue$F[["nplct"]]) * (1 + I[["ct"]]) + 
            (tissue$F[["nlct"]] + 0.3 * tissue$F[["nplct"]]) * .self$POW(tissue, chemical, I) 
    .self$P.ct <- (P.mw + tissue$F[["aplct"]] * I[["ct"]] * .self$P.aplw + tissue$F[["prct"]] * (1+I[["ct"]]) * .self$P.prwc) /
                  (1+I[["p"]])
    #cat("P.ct calculation:")
    #cat("  F.wct:  ", tissue$F[["wct"]], "\n")
    #cat("  F.nplct:", tissue$F[["nplct"]], "\n")
    #cat("  I.ct:   ", I[["ct"]], "\n") 
    #cat("  F.nlct: ", tissue$F[["nlct"]], "\n")
    #cat("  POW:    ", .self$POW(tissue, chemical, I), "\n")
    #cat("  P.mw:   ", P.mw, "\n")
    #cat("  F.aplct:", tissue$F[["aplct"]], "\n")
    #cat("  P.aplw: ", .self$P.aplw, "\n")
    #cat("  F.prct  ", tissue$F[["prct"]], "\n")
    #cat("  P.prwc: ", .self$P.prwc, "\n")
    #cat("  I.p:    ", I[["p"]], "\n")
    #cat("  P.ct:   ", .self$P.ct, "\n")
    
    return(.self$P.ct)
  },
  ##########
  ## P.it and P.p have albumin and lipoprotein, rather than the generic pr. Normally, the binding to these is a binary decision.
  ## In this case, due to my ability to calculate the fraction neutral and fraction charged at any given pH, I am de-binarifying
  ## the binding to these proteins:
  ## Old:  if(frac$neutral > 0.5) f.protein <- f$lprpp else f.protein <- f$albp
  ## New:  f.protein <- frac$neutral * tissue$F[["lprit"]] + (frac$negative + frac$positive) * tissue$F[["albit"]]
  ##########
  Pmwit = function(tissue, chemical, I)
  {
    ## Here, interstitial fluid will drop the second and third term, through zero values for nlm, nplm, and aplm.
    .self$P.aplw <- if (length(.self$P.aplw) == 0) .self$Paplw(tissue, chemical, I) else .self$P.aplw
    .self$P.prwi <- if (length(.self$P.prwi) == 0) .self$Pprwi(tissue, chemical, I) else .self$P.prwi

    frac <- chemical$FractionIonizedList(tissue$F[["pHit"]])
    ## De-binarifying the albumin/lprp data by frac neutral:
    ## Old:
    #if(frac$neutral > 0.5) f.protein <- f$lprit
    #else f.protein <- f$albit
    ## New:
    if (frac$neutral == 1) f.protein <- tissue$F[["lprit"]]
    else f.protein <- tissue$F[["albit"]]
    #f.protein <- frac$neutral * tissue$F[["lprit"]] + frac$negative * tissue$F[["albit"]]
    #P.mw <- (tissue$F[["wit"]] + 0.7 * tissue$F[["nplit"]]) * (1 + I[["it"]]) + 
    #        (tissue$F[["nlit"]] + 0.3 * tissue$F[["nplit"]]) * .self$POW(tissue, chemical, I) 
    #.self$P.it <- (P.mw + tissue$F[["aplit"]] * I[["it"]] * .self$P.aplw + f.protein * (1 + I[["it"]]) * .self$P.prwi) /
    #              (1 + I[["p"]])
    .self$P.it <- (tissue$F[["wit"]]+f.protein*.self$P.prwi)

    #if (chemical$ID$CASRN == "37517-30-9" & tissue$tissueType == "xAdipose") {
    #  cat("f.wmi (0.89):", tissue$F[["wit"]], "\n")
    #  cat("f.wp??", tissue$F[["wp"]], "\n")
    #  cat("f.lpri (0.00028981):", tissue$F[["lprit"]], "\n")
    #  cat("f.albi (0.01514765):", tissue$F[["albit"]], "\n")
    #  cat("f.protein:", f.protein, "\n")
    #  cat("P.prwi", .self$P.prwi, "\n")
    #  browser()
    #}

    #cat('.self$P.it <- (P.mw + tissue$F[["aplit"]] * I[["it"]] * .self$P.aplw + f.protein * (1 + I[["it"]]) * .self$P.prwi) / (1 + I[["p"]])\n')
    #cat('P.it <- (', P.mw, '+', tissue$F[["aplit"]], '*', I[["it"]], '*', .self$P.aplw, '+', f.protein, '* (1 +', I[["it"]], ') *', .self$P.prwi, ') / (1 + ', I[["p"]], ')\n---------------\n')
    return(.self$P.it)
  },
  Pmwp = function(tissue, chemical, I)
  {
    ## Here, interstitial fluid will drop the second and third term, through zero values for nlm, nplm, and aplm.
    .self$P.aplw <- if (length(.self$P.aplw) == 0) .self$Paplw(tissue, chemical, I) else .self$P.aplw
    .self$P.prwp <- if (length(.self$P.prwp) == 0) .self$Pprwp(tissue, chemical, I) else .self$P.prwp

    frac <- chemical$FractionIonizedList(tissue$F[["pHp"]])
    if (frac$neutral == 1) f.protein <- tissue$F[["lprp"]]
    else f.protein <- tissue$F[["albp"]]
    #f.protein <- frac$neutral * tissue$F[["lprp"]] + frac$negative * tissue$F[["albp"]]
    P.mw <- (tissue$F[["wp"]] + 0.7 * tissue$F[["nplp"]]) * (1 + I[["p"]]) + 
            (tissue$F[["nlp"]] + 0.3 * tissue$F[["nplp"]]) * .self$POW(tissue, chemical, I) 
    .self$P.p <- (P.mw + tissue$F[["aplp"]] * I[["p"]] * .self$P.aplw + f.protein * (1+I[["p"]]) * .self$P.prwp) /
                 (1 + I[["p"]])

    #if (chemical$ID$CASRN == "26675-46-7" & tissue$tissueType == "xLiver") {
    #  cat("P.mwp data:\n")
    #  cat("P.prwp:", .self$P.prwp, "\n")
    #  cat("P.aplw:", .self$P.aplw, "\n")
    #  browser()
    #}

    return(.self$P.p)
  },
  Pmwe = function(tissue, chemical, I)
  {
    ## I would postulate that the I.e would appear in th PHbw routine, and since Peyret pre-calculated the value, it cannot be un-separated.
    .self$P.aplw <- if (length(.self$P.aplw) == 0) .self$Paplw(tissue, chemical, I) else .self$P.aplw
    .self$P.prwe <- if (length(.self$P.prwe) == 0) .self$Pprwe(tissue, chemical, I) else .self$P.prwe

    P.mw <- (tissue$F[["we"]] + 0.7 * tissue$F[["nple"]]) * (1 + I[["e"]]) + 
            (tissue$F[["nle"]] + 0.3 * tissue$F[["nple"]]) * .self$POW(tissue, chemical, I) 
    ## Separated this from P.mw.ct because of the mysterious multiplication by 1, right 
    ## here ---------------------------------v
    .self$P.e <- (P.mw + tissue$F[["aple"]] * I[["e"]] * .self$P.aplw + tissue$F[["pre"]] * (1) * .self$P.prwe) /
                 (1 + I[["p"]])

    #if (chemical$ID$CASRN == "2921-88-2x") {
    #  cat('.self$P.e <- (P.mw + tissue$F[["aple"]] * I[["e"]] * .self$P.aplw + tissue$F[["pre"]] * (1) * .self$P.prwe) / (1 + I[["p"]])\n')
    #  cat('P.e <- (', P.mw, '+', tissue$F[["aple"]], '*', I[["e"]], '*', .self$P.aplw, '+', tissue$F[["pre"]], '* (1) *', .self$P.prwe, ') / (1 + ', I[["p"]], ')\n---------------\n')
    #}
    return(.self$P.e)
  },
  #################### Protein ####################
  ##########
  ## Calculate the protein:plasma (or interstitial fluid) PC
  ## As above, replacing chemical class with fractions.
  ##  if (chemical.class == "MSB" || chemical.class == "DSB" || chemical.class == "BZ") return(0)
  ##  if (chemical.class == "N") denom <- f$lprp
  ##  else denom <- f$albp
  ##########
  Pprwp = function(tissue, chemical, I)
  {
    frac <- chemical$FractionIonizedList(tissue$F[["pHp"]])
    ## Basic -> 0
    if (frac$positive > 0) {
      .self$P.prwp <- 0
    } else {
      if (frac$neutral == 1) denom <- tissue$F[["lprp"]]
      else denom <- tissue$F[["albp"]]
      #denom <- frac$neutral * tissue$F[["lprp"]] + frac$negative * tissue$F[["albp"]]
      if (denom == 0) {
        .self$P.prwp <- 0
      } else {
        F.nl <- tissue$F[["nlp"]] + 0.3 * tissue$F[["nplp"]]
        .self$P.prwp <- ((1 / chemical$fub) - 1 - (.self$POW(tissue, chemical, I) * F.nl / (1 + I[["p"]]))) / denom
        #if (chemical$ID$CASRN == "x2921-88-2") {
        #  cat("fub:", chemical$fub, "\n")
        #  cat("P.ow:", .self$POW(tissue, chemical, I), "\n")
        #  cat("F.nl:", F.nl, "\n")
        #  cat("I.p:", I[["p"]], "\n")
        #  cat("denom:", denom, "\n")
        #  cat("part 1: (1/fub)-1-P.ow*F.nl/(1+I.p):", ((1 / chemical$fub) - 1 - (.self$POW(tissue, chemical, I) * F.nl / (1 + I[["p"]]))), "\n")
        #  cat("                                   =", (1 / chemical$fub),  "- 1 -", .self$POW(tissue, chemical, I) * F.nl, "/", (1 + I[["p"]]), "\n")
        #  #browser()
        #}
      }
    }

    if(.self$P.prwp < 0) {
      .self$P.prwp <- 0
    }

    return(.self$P.prwp)
  },
  ##########
  ## P.prwi, relevant passage from Peyret, Poulin and Krishnan, Tox and Applied Pharmacology 249:197--207:
  ##   "The same value of Pprw was used for plasma and interstitial fluid since it was assumed that 
  ##    the binding macromolecules are the same in both matrices, which should represent a similar 
  ##    binding affinity and capacity (Rodgers and Rowland, 2006). The underlying principle is that 
  ##    the plasma and interstitial fluid contain these binding macromolecules that provide high-
  ##    affinity binding sites for drugs."
  ##########
  Pprwi = function(tissue, chemical, I) 
  {
    .self$P.prwi <- if(length(.self$P.prwi) == 0) .self$Pprwp(tissue, chemical, I) else .self$P.prwi
    return(.self$P.prwi)
  },
  ##########
  ## P.prwc, relevant passage from Peyret, Poulin and Krishnan, Tox and Applied Pharmacology 249:197--207:
  ##   "It is known that for small chemical molecules, distribution is typically driven by nonspecific 
  ##    binding in tissues; however, pharmacological target binding is generally of minor relevance in 
  ##    terms of their contribution to the volume of distribution. Therefore, the binding to intra-
  ##    cellular proteins has not been characterised routinely in efforts focusing ot predict PCs. In 
  ##    other words, the extent of binding to the proteins in tissue cellular matrix was set equal to 
  ##    zero for all compounds."
  ##########
  Pprwc = function(tissue, chemical, I) 
  {
    .self$P.prwc <- 0
    return(.self$P.prwc)
  },
  ## Same as above.
  Pprwe = function(tissue, chemical, I) 
  {
    #if (chemical$ID$CASRN == "26675-46-7" & tissue$tissueType == "xLiver") browser()
    .self$P.prwe <- if (is.VOC(chemical$vapor.pressure)) PHbw(tissue, chemical, I) else 0
    return(.self$P.prwe)
  },
  #################### Acidic Phospholipids ####################
  ##########
  ## Calculate the acidic phospholipid:water partition coefficient.
  ##########
  Paplw = function(tissue, chemical, I) 
  {
    if (length(chemical$Kpu.BC) == 0) {
      chemical$Kpu.BC <- f.Kpu.BC(tissue, chemical, I)
    }
    ## Determination of chemical class of compound.
    frac <- chemical$FractionIonizedList(tissue$F[["pHct"]])
    if (chemical$Kpu.BC < 0 || frac$negative > 0 || frac$neutral == 1) {    ## Only calculate for basic compounds. ##
      .self$P.aplw <- 0
    } else if (I[["e"]] == 0) {
      ## Denominator is I.e * f.aple -- P.aplw <- Inf.
      .self$P.aplw <- 0
    } else {
      numerator <- (1 + I[["e"]]) * (tissue$F[["we"]] + 0.7 * tissue$F[["nple"]]) + 
                   (.self$POW(tissue, chemical, I) * (tissue$F[["nle"]] + 0.3 * tissue$F[["nple"]]))
      ## Explanation: Since the assumption that ionization occurs only in the water phase, then a certain amount of the chemical CAN be
      ## partitioned into the aplw, at which point, no more ionization/deionization takes place. Thus, we assume the equilibrium negative
      ## fraction can be partitioned into the apl
      .self$P.aplw <- ( chemical$Kpu.BC - (numerator / (1 + I[["p"]]))) * ((1 + I[["p"]]) / (I[["e"]] * tissue$F[["aple"]]))
      if (.self$P.aplw < 0) .self$P.aplw <- 0
      #else .self$P.aplw <- .self$P.aplw - frac$negative * .self$P.aplw     ## I think this is more correct, but whatever. ##
    }

    #if (chemical$ID$CASRN == "72956-09-3" & tissue$tissueType == "xLiver") {
    #  cat("Kpu.BC:", chemical$Kpu.BC, "\n")
    #  cat("P.ow:", .self$POW(tissue, chemical, I),"\n")
    #  cat("f.we:", tissue$F[["we"]], "\n")
    #  cat("f.nple:", tissue$F[["nple"]], "\n")
    #  cat("I.p:", I[["p"]], "\n")
    #  cat("I.e", I[["e"]], "\n")
    #  #browser()
    #}

    return(.self$P.aplw)
  },
  #################### Partition coefficient, unbound, Blood Cells ####################
  ##########
  ## Calculate the erythrocyte:plasma PC
  ##########
  f.Kpu.BC = function(tissue, chemical, I)    # B.P set tu unity for chemicals in which data does not exist, as in Poulin & Haddad, 2012
  {
    f.e <- 0.45
    #f.e <- tissue$F[["e"]]
    Kpu.BC <- ((chemical$B.P - (1 - f.e)) / f.e) / chemical$fub
    return(Kpu.BC)
  },
  ##########
  ## Volatility of organic chemicals needed for:
  ## - Calculation of P.Hbw
  ## - Determination of treatement of highly lipophilic organic compounds (Haddad, Poulin, & Krishnan, 2000; Chemosphere)
  ##########
  is.VOC = function(vapor.pressure, boiling.point = NULL)
  {
    isvoc <- data.frame(by.john.little = FALSE, by.regulatory.agencies = TRUE, stringsAsFactors=FALSE)
    ## John Little definition
    if (vapor.pressure > 10) {
      ### HBW only for VOC & VVOC
      isvoc$by.john.little <- TRUE    ## VOC
    } else if (vapor.pressure > 0.000000001) {
      ## NOT FOR SVOC
      isvoc$by.john.little <- TRUE
    } else {
      isvoc$by.john.little <- FALSE
    }

    if (length(boiling.point) == 0) {
      isvoc$by.regulatory.agencies <- FALSE
    } else if(boiling.point < 250) { 
      ## EU/Canada/EPA definition
      isvoc$by.regulatory.agencies <- TRUE 
    }

    return(isvoc$by.regulatory.agencies)
  },
  ##########
  ## A general check for the HLOC'ness of a particular chemical
  ##########
  is.HLOC = function(tissue, chemical, I) {
    if (chemical$is.HLOC) return(TRUE)
    else return(FALSE)
    ## Haddad & Krishnan, Chemosphere 40:839-843, 2000
#    if (chemical$logP > 4) {    
      ## Poulin & Haddad, J. Pharm. Sci. 101:2250-2261
      if (chemical$logP > 5.8) {    
        # Kp.HLOC
        return(TRUE)
      }
      ## Applies only to adipose tissue.
#      if (tissue$tissueType == "Adipose") { 
        # Patb.HLOC 
#        return(TRUE) 
#      }
#    } 
    return(FALSE)
  },
  ##########
  ## Gives the appropriate P.ow/P.vow
  ##########
  POW = function(tissue, chemical, I) {
    ## In the verification, Rodgers and Rowland (2005) used P.vow for Liver. In Rodgers and Rowland (2006), used P.ow.
    ## Here, we are using P.ow to agree with the more recent research.
    if (tissue$tissueType == "Adipose") {
      return(chemical$P.vow)
    } else {
      return(chemical$P.ow)
    }
  }
)

