##########
# QSARs.R
# Author: Cory L. Strope
# Date: Jan 16, 2015
#
# Description: 
# 
##########

##########
### The conversion from P.ow to P.vow. Not sure which conversion to use, so all are included, just commented.
##########
Pvow <- function(P.ow, SMILES = NULL, chemical.class = NULL) 
{
  #####
  ## There are a number of different transformations of P.ow to P.vow. Suggested techniques for doing the appropriate transformation is taken from:
  ##
  ## log.P.vow <- 1.012*log10(P.ow)+0.0932    ## Poulin & Krishnan, 1996
  ## log.P.vow <- 1.119*log10(P.ow)-0.325     ## Leo, Hansch & Elkins, 1971
  ## log.P.vow <- 1.115*log10(P.ow)-1.34      ## SimulationsPlus
  #####

  log.P.vow <- 1.115*log10(P.ow)-1.34      ## SimulationsPlus
  if(!is.null(SMILES)) {
    if( any(SMILES == "O") || any(SMILES == "o") ) {
      log.P.vow <- 1.099*log10(P.ow)-1.31      ## Poulin & Haddad, 2012; contains at least 1 oxygen atom in the molecular structure
      # cat("log.P.vow -- oxygen detected:", log.P.vow, "\n")
    } else if (!is.null(chemical.class)) {
      if( any("S", chemical.class) ) {   ## Because only bases are classified as weak (W) or strong (S), checking for S is sufficient.
        log.P.vow <- 1.0654*log10(P.ow)-0.232    ## Poulin & Haddad, 2012; contains no oxygen atoms, or is a strong base.  Average of first 2 equations above.
        # cat("log.P.vow -- chemical class strong base:", log.P.vow, "\n")
      }
    }
  }
  #cat("log.P.vow post:", log.P.vow, "\n")
  P.vow <- 10^(log.P.vow)
  #cat("P.vow: ", P.vow, "\n")
  return(P.vow)
}

##########
## The QSAR for calculating the blood to plasma ratio via the Buist (2012) method. 
## This methods fit-for-purpose range appears to be *only* pharma compounds (primarily to the restricted logP range for Rx).
##########
Buist2012_BP <- function(chemical) {
  B.P <- 10^(-0.004282 + 0.067028*chemical$logP + 0.214590*log10(chemical$fub))
  return(B.P)
}

##########
## Calculates the human partition of blood air according to Buist2012. 
##########
Buist2012_Pba <- function(chemical) {
  if (length(chemical$blood.air) > 0) {
    return(chemical$blood.air)
  }
  if (length(chemical$vapor.pressure) == 0) {
    chemical$vapor.pressure <- 1
    message("Warning for blood.air partition coefficient.\n  Chemical ", chemical$ID$Compound, " is missing vapor pressure. Assuming vapor.pressure == 1.\n")
  }
  Pba <- 6.96 - 1.04 * chemical$vapor.pressure - 0.533 * chemical$logP - 0.00495 * chemical$MW
  return(Pba)
}