##########
# normalizeInput.R
# Author: Cory L. Strope
# Date: Jan 16, 2015
#
# Description: 
# This file contains routines that accept data frames from pKa prediction methods and transforms the data into the format
# that is usable by pKaTools. Current predictors with methods of transformation are SPARC, ADMET Predictor 7.0 (some inference
# methods were incorporated, likely incorrectly inferred acid/base pKas from reported mixed pKa), and pKa PlugIn from ChemAxon.
##########

#####
## This function changes the various input data frames for predictors into an internal format that is used
## only within pKaTools. Much easier to deal with. (may need to change back after analysis?)
#####
normalize_pKa <- function(pKa_df) {
  #####
  ## Start easy, replace column names that have standard names.
  # Acids #
  # Make replacements of the easiest first, (e.g., apKa vs pKa), then do easier ones.
  colnames(pKa_df) <- str_replace_all(colnames(pKa_df), "S.Acidic_pKa", "PKATOOLS_APKA")  # ADMET Pred. R reads in '+' as '.'
  colnames(pKa_df) <- str_replace_all(colnames(pKa_df), "S.Basic_pKa", "PKATOOLS_BPKA")
  colnames(pKa_df) <- str_replace_all(colnames(pKa_df), "S.Mixed_pKa", "PKATOOLS_MPKA")
  colnames(pKa_df) <- str_replace_all(colnames(pKa_df), "apKa|Acid", "PKATOOLS_APKA") # MARVIN (and I thought sparc for the ACID...)
  colnames(pKa_df) <- str_replace_all(colnames(pKa_df), "bpKa|Base", "PKATOOLS_BPKA")
  # Bases #
  colnames(pKa_df) <- str_replace_all(colnames(pKa_df), "pKa", "PKATOOLS_APKA")   # SPARC, & MARVIN x.atoms
  colnames(pKa_df) <- str_replace_all(colnames(pKa_df), "pKb", "PKATOOLS_BPKA")   # SPARC

  #####
  ## All necessary columns should have the prefix "PKATOOLS". Any columns that are not pKaTools can now be placed
  ## into a normalized data frame. Then, add columns in the way that I want them to be normal.
  keeps <- which(grepl("PKATOOLS", colnames(pKa_df)) == FALSE)  ## Grab untampered with columns
  normalized_pKa_df <- pKa_df[,keeps]
  ## Now, adding in the pKa data. FORMAT: 4 columns, pKaTools_APKA (no number) pKaTools_APKA.atoms (no number).
  ## Internal to columns is a semicolon separated list of the pKa values.
  # Cases:
  # (1) MARVIN: 3 pKaTools APKA & BPKA, each with either nothing or a single value. 1 pKaTools_X.atoms, comma separated
  # (2) SPARC: 3 pKaTools APKA, BPKA (nothing|1pKa), APKA.atoms, and BPKA.atoms (csv for multiple atoms).
  # (3) ADMET Pred: APKA, BPKA, and "mixed" (what to do??). Multiple pKa predictions, separated by " ;", or NONE (remove).
  nchem <- nrow(normalized_pKa_df)
  pKaTools_df <- data.frame(pKaTools_APKA=rep(NA,nchem), pKaTools_APKA.atoms=rep(NA,nchem), 
                pKaTools_BPKA=rep(NA,nchem), pKaTools_BPKA.atoms=rep(NA,nchem))
  format_type <- pKa_normalization_type(colnames(pKa_df))
  #cat("Formatting pKa (", format_type, ") to pKaTools format.\n", sep="")
  if (nrow(pKa_df) > 1) pb <- txtProgressBar(min=1, max=nchem, initial=1, style=3)
  LIST_START <- 1   ## To check specific errors. Remove this when all ok.
  LIST_END <- nchem
  for (i in LIST_START:LIST_END) {
    if (nrow(pKa_df) > 1) setTxtProgressBar(pb, i) 
    #cat("******************************", i, pKa_df[i,1], pKa_df[i,2], "*************************************\n")
    normalized_row <- normalize_row(pKa_df[i,], format_type)
    pKaTools_df$pKaTools_APKA[i] <- normalized_row$pKaTools_APKA
    pKaTools_df$pKaTools_BPKA[i] <- normalized_row$pKaTools_BPKA
    pKaTools_df$pKaTools_APKA.atoms[i] <- normalized_row$pKaTools_APKA.atoms
    pKaTools_df$pKaTools_BPKA.atoms[i] <- normalized_row$pKaTools_BPKA.atoms
  } 
  cat("\n") ## Add endline so that next progressbar is on the newline.
  #print(pKaTools_df)
  normalized_pKa_df <- cbind(normalized_pKa_df, pKaTools_df)

  return(normalized_pKa_df)
}

#####
## Primary function for the conversion of pKa input to a standardized format the is useful for the package.
#####
normalize_row <- function(chemRow, format_type) {
  row_df <- data.frame(pKaTools_APKA="", pKaTools_APKA.atoms="", pKaTools_BPKA="", pKaTools_BPKA.atoms="", stringsAsFactors=FALSE)
  if (format_type == "MARVIN-like") {
    ## concatenate all of the pKas together, separated by ';'
    # ACID
    apKa.tmp <- unlist(chemRow[grepl("PKATOOLS_APKA\\d", colnames(chemRow))])
    if (length(apKa.tmp) != 0) apKa.tmp <- apKa.tmp[!is.na(apKa.tmp)]
    row_df$pKaTools_APKA <- paste(apKa.tmp, collapse=";")
    # BASE
    bpKa.tmp <- unlist(chemRow[grepl("PKATOOLS_BPKA\\d", colnames(chemRow))])
    if (length(bpKa.tmp) != 0) bpKa.tmp <- bpKa.tmp[!is.na(bpKa.tmp)]
    row_df$pKaTools_BPKA <- paste(bpKa.tmp, collapse=";")
    ## Atoms are a comma separated list in MARVIN
    if (row_df$pKaTools_APKA == "") {
      row_df$pKaTools_APKA.atoms <- ""
    } else {
      ## If MARVIN-like, but no positions given, put only zeroes.
      row_df$pKaTools_APKA.atoms <- gsub(",", ";", chemRow[grepl("PKATOOLS_APKA.atoms", colnames(chemRow))])
      ## if row_df$pKaTools_APKA.atoms was NA in input, the the returned value from above is "NA"
      if (row_df$pKaTools_APKA.atoms == "NA") {
        row_df$pKaTools_APKA.atoms <- paste(rep(0,length(apKa.tmp)), collapse=";")   ## If length=0, this is ""
      }
    }
    if (row_df$pKaTools_BPKA == "") {
      row_df$pKaTools_BPKA.atoms <- ""
    } else {
      row_df$pKaTools_BPKA.atoms <- gsub(",", ";", chemRow[grepl("PKATOOLS_BPKA.atoms", colnames(chemRow))])
      if (row_df$pKaTools_BPKA.atoms == "NA") {
        row_df$pKaTools_BPKA.atoms <- paste(rep(0,length(bpKa.tmp)), collapse=";")   ## If length=0, this is ""
      }
    }
  } else if (format_type == "SPARC-like") {
    ## SPARC HAS:
    # PKATOOLS_APKA1   PKATOOLS_APKA1.atom
    # PKATOOLS_APKA2   PKATOOLS_APKA2.atom
    # PKATOOLS_APKA3   PKATOOLS_APKA4.atom
    # PKATOOLS_BPKA1   PKATOOLS_BPKA1.atom
    # PKATOOLS_BPKA2   PKATOOLS_BPKA2.atom
    # PKATOOLS_BPKA3   PKATOOLS_BPKA3.atom
    ## And 2 error messages:
    # Calc.error     Calculation.Time.Too.Long
    ## We will assume that these mean no predicted pKas; SPARC is useless for these data.
    #cat("xxxx:", chemRow$PKATOOLS_APKA1, "\n")
    if(!is.null(chemRow$PKATOOLS_APKA1)) 
      if (chemRow$PKATOOLS_APKA1 == "Calc Error"                        ## hack hack... cough... Need heimlich!
        | chemRow$PKATOOLS_APKA1 == "Calculation Time Too Long"         ## SPARC slow. SMILES string long. SPARC no like.
        | chemRow$PKATOOLS_APKA1 == "skipped"                           ## Usually accompanied with the grepl below
        | chemRow$PKATOOLS_APKA1.atom == "skipped"                      ## ?? Why it is in .atoms I cannot know.
        | chemRow$PKATOOLS_APKA1.atom == "Bad SMILES string for SPARC"  ## I have no idea why SPARC cannot accommodate
        | grepl("SPARC calculations", chemRow$PKATOOLS_APKA1.atom)      ## SPARC takes too many calculations
      ) 
        return(row_df)

    ## Passed initial check. Reset to nothing.
    row_df <- data.frame(stringsAsFactors=FALSE)
    ## Herein lies other garbagey items introduced into SPARC. Need to deal with comma-separated lists AND 
    ## csl's separated by '|'.
    #cat("Check wtf row_df is for each of these:\n")
    #browser()
    row_df.acid <- SPARC2pKaTools(
                                  chemRow[grepl("PKATOOLS_APKA\\d$", colnames(chemRow))],
                                  chemRow[grepl("PKATOOLS_APKA\\d.atom", colnames(chemRow))]
                                 )
    row_df.base <- SPARC2pKaTools(
                                  chemRow[grepl("PKATOOLS_BPKA\\d$", colnames(chemRow))],
                                  chemRow[grepl("PKATOOLS_BPKA\\d.atoms", colnames(chemRow))]
                                 )
    ## Now, paste together all of the row_df data into PKATOOLS format. ##
    ## gsub: if either apka or bpka have uneven numbers, the remainder of apka or bpka are filled with "".
    ## Upon collapse (to make new row_df), these empty strings will be separated by ';', which is not in the 
    ## pKaTools specs. Thus, replace all "" with NA, then upon collapse, remove all NA's.
    apKa.vec <- row_df.acid$pKaTools_APKA
    apKa.atoms <- row_df.acid$pKaTools_APKA.atoms
    bpKa.vec <- row_df.base$pKaTools_BPKA
    bpKa.atoms <- row_df.base$pKaTools_BPKA.atoms
    row_df <- data.frame(
                         pKaTools_APKA = paste(apKa.vec[apKa.vec != ""], collapse=";"),
                         pKaTools_APKA.atoms = paste(apKa.atoms[apKa.atoms != ""], collapse=";"),
                         pKaTools_BPKA = paste(bpKa.vec[bpKa.vec != ""], collapse=";"),
                         pKaTools_BPKA.atoms = paste(bpKa.atoms[bpKa.atoms != ""], collapse=";"),
                         stringsAsFactors=FALSE
                        )
  } else if (format_type == "ADMETPred-like") {
    ## Deal with the mixed pKa... Add to both Acid and Base (similar to SPARC making equal pKa predictions for both Acid and Base.
    # ACID
    row_df <- ADMETPred2pKaTools(chemRow)
  } else { 
    stop("Unrecognizable pKa input format.\n")
  }
  
  return(row_df)
}

#####
## This won't fix the NAs, but at least it should reduce the code.
#####
append_pKa <- function(append.where, append.what, gt0) {
  if (gt0 == 0) return(append.what)
  else return(paste0(append.where, ";", append.what))
}

#####
## Another function to reduce the hugeness and confusingness of ADMETPred2pKaTools
#####
get_pKa_from_ADMETPred_input <- function(chem.df, togrep) {
  ## Each search of togrep, in ADMETPred, will return exactly 1 item. Split that item ([[1]]) to get the current pKas.
  pKa.tmp <- unlist(strsplit((chem.df[grepl(togrep, colnames(chem.df))])[[1]], ";\\s*"))
  if (length(pKa.tmp) == 1) if(pKa.tmp == "None" | pKa.tmp == "") pKa.tmp <- NULL
  pKa.tmp <- as.numeric(unlist(pKa.tmp))
  pKaTools_PKA <- paste0(pKa.tmp, collapse=";")
  # No placement info, so create a temporary vector that puts the correct number of zeroes.
  pKaTools_PKA.atoms <- paste(rep(0,length(pKa.tmp)), collapse=";")   ## If length=0, this is ""

  ## Remove any whitespace. Not sure if necessary, but doesn't hurt, can only help.
  if (!is.null(pKaTools_PKA.atoms)) pKaTools_PKA.atoms  <- gsub("\\s+", "", pKaTools_PKA.atoms)
  if (!is.null(pKaTools_PKA)) pKaTools_PKA  <- gsub("\\s+", "", pKaTools_PKA)

  pKaTools_pKa.df <- data.frame(pKa=pKaTools_PKA, atoms=pKaTools_PKA.atoms, stringsAsFactors=FALSE)
  return(pKaTools_pKa.df)
}

#####
## Mixed pKas <- pain_in_neck.
#####
ADMETPred2pKaTools <- function(chemRow) {
  ## Set all data to row_df. Unfortunately, mixed pKa present a rather more difficult problem, since it is a chemical-specific
  ## data structure, rather than something I would prefer, i.e., atom-specific pKa bindings.
  ## These are data.frames with 2 columns: pKa and atoms, each in the proper pKaTools format.
  apKa.df <- get_pKa_from_ADMETPred_input(chemRow, "PKATOOLS_APKA")
  bpKa.df <- get_pKa_from_ADMETPred_input(chemRow, "PKATOOLS_BPKA")
  mpKa.df <- get_pKa_from_ADMETPred_input(chemRow, "PKATOOLS_MPKA")
  ## Get the vectors of values to work through the mpKa-a/bpKa operations.
  ## * can do as.numeric since the "None" values from ADMETPred are removed by the above df calls.
  apKa.vec <- as.numeric(unlist(strsplit(apKa.df$pKa, ";")))
  bpKa.vec <- as.numeric(unlist(strsplit(bpKa.df$pKa, ";")))
  mpKa.vec <- as.numeric(unlist(strsplit(mpKa.df$pKa, ";")))

  ## If fingerprint file exists, try and infer (to some extent). If not, then apply simple-minded stock rules that will
  ## do moderately well, I think.
  ## Here, doing a relatively simplistic sorting of the pKa values reported by ADMETPred. See below comments for explanations:
  if (length(mpKa.vec) == 1) {
    #cat("SINGLE mpKa\n")
    ## Assume charge of atom is the OPPOSITE of the closest pKa in APKA or BPKA
    ## Find closest acidic pKa by taking the minimum of the absolute value of differences between apKa and mpKa
    if (length(apKa.vec) == 0) apKa_ <- 0
    else apKa_ <- min(abs(apKa.vec - mpKa.vec))

    ## Find closest acidic pKa by taking the minimum of the absolute value of differences between apKa and mpKa
    if (length(bpKa.vec) == 0) bpKa_ <- 0
    else bpKa_ <- min(abs(bpKa.vec - mpKa.vec))

    if (apKa_ < bpKa_) {
      apKa.df$pKa <- append_pKa(apKa.df$pKa, mpKa.vec, if (apKa.df$pKa=="") 0 else 1)
      apKa.df$atoms <- append_pKa(apKa.df$atoms, "0", if (apKa.df$atoms=="") 0 else 1)
    } else {
      bpKa.df$pKa <- append_pKa(bpKa.df$pKa, mpKa.vec, if (bpKa.df$pKa=="") 0 else 1)
      bpKa.df$atoms <- append_pKa(bpKa.df$atoms, "0", if (bpKa.df$atoms=="") 0 else 1)
    }
  } else if (length(mpKa.vec) == 2) {
    #cat("Double mpKa\n")
    ## 2 cases
    ##   IF both mixed pKa are within ~1.5, assume that these two confound each other
    ##   IF there is a decent separation between the two pKa's, make assumption similar to above
    apKa_ <- vector()
    bpKa_ <- vector()
    for (i in 1:length(mpKa.vec)) {
      ## Find closest acidic pKa by taking the minimum of the absolute value of differences between apKa and mpKa
      if (length(apKa.vec) == 0) apKa_[i] <- 0
      else apKa_[i] <- min(abs(apKa.vec - mpKa.vec[i]))
      
      ## Find closest acidic pKa by taking the minimum of the absolute value of differences between apKa and mpKa
      if (length(bpKa.vec) == 0) bpKa_[i] <- 0
      else bpKa_[i] <- min(abs(bpKa.vec - mpKa.vec[i]))

      if (apKa_[i] == 0 & bpKa_[i] == 0) { ## Only mixed pKa. Assume to be basic?
        bpKa_[i] <- -mpKa.vec[i]    ## This will be guaranteed to be lower than apKa. (which is 0)
      }
    }

    #### Making this super-simple. Mixed pKa interact with each other for even number of mixed pKa, and for odd
    #### numbers, two interact, one is checked against a/bpKa.
    ## Have mpKa to a/bpKa differences. Need to also evaluate m/mpKa differences. Only issue with this is that I
    ## will not know which is the acidic or which is the basic.........
    ## Set each mixed pKa as above (opposite charge as closest relative.)

    if (is.null(apKa_[1]) | is.na(apKa_[1])) apKa_[1] <- 0
    if (is.null(apKa_[2]) | is.na(apKa_[2])) apKa_[2] <- 0
    if (is.null(bpKa_[1]) | is.na(bpKa_[1])) bpKa_[1] <- 0
    if (is.null(bpKa_[2]) | is.na(bpKa_[2])) bpKa_[2] <- 0

    for (i in 1:length(mpKa.vec)) {
      if (apKa_[i] < bpKa_[i]) {
        apKa.df$pKa <- append_pKa(apKa.df$pKa, mpKa.vec[i], if (apKa.df$pKa=="") 0 else 1)
        apKa.df$atoms <- append_pKa(apKa.df$atoms, "0", if (apKa.df$atoms=="") 0 else 1)
      } else {
        bpKa.df$pKa <- append_pKa(bpKa.df$pKa, mpKa.vec[i], if (bpKa.df$pKa=="") 0 else 1)
        bpKa.df$atoms <- append_pKa(bpKa.df$atoms, "0", if (bpKa.df$atoms=="") 0 else 1)
      }
    }
  } else if (length(mpKa.vec) > 2) {
    #cat("****************************************", length(mpKa.vec), "mpKas\n")
    ## ? No idear. Probably a mixture of the above 2 cases. Do best you can. ##
    apKa_ <- vector()
    bpKa_ <- vector()
    apKa__ <- vector()
    bpKa__ <- vector()
    for (i in 1:length(mpKa.vec)) {
      ## Find closest acidic pKa by taking the minimum of the absolute value of differences between apKa and mpKa
      if (length(apKa.vec) == 0) apKa_[i] <- 0
      else apKa_[i] <- min(abs(apKa.vec - mpKa.vec[i]))

      ## Find closest acidic pKa by taking the minimum of the absolute value of differences between apKa and mpKa
      if (length(bpKa.vec) == 0) bpKa_[i] <- 0
      else bpKa_[i] <- min(abs(bpKa.vec - mpKa.vec[i]))
      
      if (apKa_[i] == 0 & bpKa_[i] == 0) { ## Only mixed pKa. Assume to be basic?
        bpKa_[i] <- -mpKa.vec[i]    ## This will be guaranteed to be lower than apKa. (which is 0)
      }    
    }

    ## Find 2 closest mixed pKa:
    r.mpKa <- mpKa.vec
    min.pair <- c(0,0)

    while (length(r.mpKa) > 1) {
      min_mixed <- 10000
      for (i in 1:(length(r.mpKa)-1)) {
        for (j in (i+1):length(r.mpKa)) {
          if (abs(r.mpKa[i]-r.mpKa[j]) < min_mixed) {
            min_mixed <- abs(r.mpKa[i]-r.mpKa[j])
            min.pair <- c(i,j)
          }
        }
      }

      #cat("\nPair being sorted: ")
      #print(r.mpKa[c(i,j)])
      #cat("\n")

      ## Get log difference between closest a/bpKa (for sorting to apKa or bpKa)
      if (length(apKa.vec) == 0) apKa__[1] <- 0
      else apKa__[1] <- min(abs(apKa.vec - min.pair[1]))
      if (length(bpKa.vec) == 0) bpKa__[1] <- 0
      else bpKa__[1] <- min(abs(bpKa.vec - min.pair[1]))

      if (apKa__[1] < bpKa__[1]) {
        ## If(x) 0 else 1... either append 0 values or 1 value (length to pass to append_pKa.
        apKa.df$pKa <- append_pKa(apKa.df$pKa, r.mpKa[min.pair[1]], if(apKa.df$pKa == "") 0 else 1)
        apKa.df$atoms <- append_pKa(apKa.df$atoms, "0", if(apKa.df$atoms == "") 0 else 1)
        bpKa.df$pKa <- append_pKa(bpKa.df$pKa, r.mpKa[min.pair[2]], if(bpKa.df$pKa == "") 0 else 1)
        bpKa.df$atoms <- append_pKa(bpKa.df$atoms, "0", if(bpKa.df$atoms == "") 0 else 1)
      } else {
        bpKa.df$pKa <- append_pKa(bpKa.df$pKa, r.mpKa[min.pair[1]], if(bpKa.df$pKa == "") 0 else 1)
        bpKa.df$atoms <- append_pKa(bpKa.df$atoms, "0", if(bpKa.df$atoms == "") 0 else 1)
        apKa.df$pKa <- append_pKa(apKa.df$pKa, r.mpKa[min.pair[2]], if(apKa.df$pKa == "") 0 else 1)
        apKa.df$atoms <- append_pKa(apKa.df$atoms, "0", if(apKa.df$atoms == "") 0 else 1)
      }

      ## Remove the sorted pair from consideration.
      r.mpKa <- r.mpKa[-min.pair]
    }

    ## If there is yet another left-over, sort it accordingly.
    if (length(r.mpKa) == 1) {
      ## Does doing in an apply take into account if a/bpKa is null? Hope so...
      if (length(apKa.vec) == 0) a <- 0
      else a <- min(abs(apKa.vec - r.mpKa))
      if (length(bpKa.vec) == 0) b <- 0
      else b <- min(abs(bpKa.vec - r.mpKa))
      if (a < b) {
        apKa.df$pKa <- append_pKa(apKa.df$pKa, r.mpKa, if(apKa.df$pKa == "") 0 else 1)
        apKa.df$atoms <- append_pKa(apKa.df$atoms, "0", if(apKa.df$pKa == "") 0 else 1)
      } else {
        bpKa.df$pKa <- append_pKa(bpKa.df$pKa, r.mpKa, if(bpKa.df$pKa == "") 0 else 1)
        bpKa.df$atoms <- append_pKa(bpKa.df$atoms, "0", if(bpKa.df$pKa == "") 0 else 1)
      }
    }
  }

  ## Now that all the data is together, add to data frame and return.
  row_df <- data.frame(pKaTools_APKA=apKa.df$pKa, 
                       pKaTools_APKA.atoms=apKa.df$atoms, 
                       pKaTools_BPKA=bpKa.df$pKa, 
                       pKaTools_BPKA.atoms=bpKa.df$atoms, 
                       stringsAsFactors=FALSE
                      )
  ## After creating this data.frame, there should be exactly 1 row; all pKa data is in pKaTools format "pka;pka;...;pka"

  #if (length(mpKa.vec) > 0) {
  #  print(row_df)
  #  cat("mpKas: ", mpKa.vec, "\n")
  #  #browser()
  #}

  return(row_df)
}

#####
## As much as I hate having to write this because of SPARC.....
## This function takes the SPARC pKa location report and properly deals with them:
## * SPARC gives a list of locations, comma-separated. This is ok, but needs to be changed.
## * SPARC also gives a list of lists, separated by a '|'. This is not ok.
#####
SPARC2pKaTools <- function(SPARC_row_pKa, SPARC_row_pKa_position) {
  #cat("**************************pKa and positions *************************\n")
  #print(SPARC_row_pKa)
  #print(SPARC_row_pKa_position)
  #cat("======================================================================\n")
  ### TODO: FOR SPARC, SEARCH EACH PKA POSITION FOR THE FIRST COMMA AND DELETE BOTH IT AND EVERYTHING FOLLOWING. ####
  PKA <- if (grepl("APKA", names(SPARC_row_pKa)[1]) == TRUE) TRUE else FALSE
  ## Remove the garbage from the SPARC row reporting.
  x <- which(is.na(SPARC_row_pKa))
  if (length(x) != 0) row.tmp <- SPARC_row_pKa[-which(is.na(SPARC_row_pKa))]
  else row.tmp <- SPARC_row_pKa
  if (length(row.tmp) > 0) {
    x <- which(row.tmp == "")
    if (length(x) != 0) row_pKa <- row.tmp[-which(row.tmp == "")]
    else row_pKa <- row.tmp
  } else row_pKa <- row.tmp     ## NULL frame...

  x <- which(is.na(SPARC_row_pKa_position))
  if (length(x) != 0) row.tmp <- SPARC_row_pKa_position[-which(is.na(SPARC_row_pKa_position))]
  else row.tmp <- SPARC_row_pKa_position
  if (length(row.tmp) > 0) {
    x <- which(row.tmp == "")
    if (length(x) != 0) row_pos <- row.tmp[-which(row.tmp == "")]
    else row_pos <- row.tmp
  } else row_pos <- row.tmp  
  if (length(row_pKa) != length(row_pos)) {
    print(SPARC_row_pKa)
    print(SPARC_row_pKa_position)
    stop("pKa predictions not aligned? Different number of reported pKas and pKa positions in SMILES strings.\n")
  }


  ## Create the data.frame to be returned ##
  pKaTools.df <- data.frame(pKaTools_XPKA = rep("", length(row_pKa)), 
                            pKaTools_XPKA.atoms = rep("", length(row_pKa)), 
                            stringsAsFactors=FALSE
                           )
  atom_positions <- list()       ## Variable to keep the SMILES position in comma-delimited format.

  if (length(row_pKa) > 0 | is.null(row_pKa)) {
    for (i in 1:length(row_pKa)) {
      ## positions are more difficult than pKa values: Values are 1 value per pKa\\d. Positions tend to be a list of potential atoms.
      #cat("row_pos$", names(row_pos)[i], sep="")
      atoms.nobar <- as.character(unlist(eval(parse(text=paste0("row_pos$", names(row_pos)[i])))))
      atom_positions[[i]] <- unlist(strsplit(atoms.nobar, ",", fixed=TRUE))
      pKaTools.df$pKaTools_XPKA[i] <- row_pKa[i]
    }

    for (i in 1:length(atom_positions)) {
      ## Determine whether the list has a possible assignment from the list of positions originally given. ##
      possible_assignments <- atom_positions[[i]]
      if (length(possible_assignments) == 0) {  ## All numbers have been previously used
        pKaTools.df$pKaTools_XPKA.atoms[i] <- 0     ## Complex attempt to try and assign. First see if this happens. ##
        cat("OOOOOOOOOOOOOOHHHHHHHHHHHHHHHHHH NNNNNNNNNNNNNNOOOOOOOOOOOOEEEEEEEESSSSSSSSSSSSSSSSSSSS!!!!!")
      } else {
        #cat(position_list, ", setting", possible_assignments[1])
        pKaTools.df$pKaTools_XPKA.atoms[i] <- possible_assignments[1]
        ## Remove all further occurrences of the SMILES position (pending some SPARCish BS) ##
        for (x in i:length(atom_positions)) {
          ## If there is only 1 element in the list, then SPARC has made a double prediction to the same position in the SMILES. ##
          if (length(atom_positions[[x]]) > 1) {
            ## More than 1 element: Remove further references to this particular SMILES position. ##
            z <- which(possible_assignments[1] %in% atom_positions[[x]])        
            ## If a position is found, then remove it ##
            if (length(z) > 0) atom_positions[[x]] <- atom_positions[[x]][-z]
          }
        }
      }
    }
  } else {
    pKaTools.df <- data.frame(pKaTools_XPKA = "", pKaTools_XPKA.atoms = "", stringsAsFactors=FALSE)
  }

  ## Above is a generic-ish data frame. Before returning, need to make it reflect what the pKas actually are. ##
  names(pKaTools.df)[1] = if(PKA) "pKaTools_APKA" else "pKaTools_BPKA"
  names(pKaTools.df)[2] = if(PKA) "pKaTools_APKA.atoms" else "pKaTools_BPKA.atoms"

  #print(pKaTools.df)
  #cat("----------------------------------------------------------------------\n")
  #cat("----------------------------------------------------------------------\n")
  return(pKaTools.df)
}

#####
## Function that tries to determine the pKa format type (per predictor input).
## This is needed for the row normalization.
#####
pKa_normalization_type <- function(column_names) {
  # number of pka columns organized by digit (MARVIN, SPARC)
  npKad_col <- length(grep("PKATOOLS_APKA\\d", column_names))
  # Total number of pKa columns (MARVIN has pKaTools_APKA.atoms, SPARC has pKaTools_APKA1.atom)
  npKaa.atoms_col <- length(grep("PKATOOLS_APKA.atoms", column_names))      # MARVIN specific
  npKad.atoms_col <- length(grep("PKATOOLS_APKA\\d.atom", column_names))  # SPARC specific
  npKa_col <- length(grep("PKATOOLS_APKA", column_names))       # ADMET Pred (no digit at end)

  if (ADMETPRED) return("ADMETPred-like")
  if (SPARC) return("SPARC-like")
  if (MARVIN) return("MARVIN-like")

  #cat("columns:", column_names, "\n", npKad_col, npKaa.atoms_col, npKad.atoms_col, npKa_col, "\n\n")
  #browser()
  if (npKaa.atoms_col == 1) {
    ## MARVIN
    return("MARVIN-like")
  } else if (npKad_col == 2*npKad.atoms_col) {    ## because PKATOOLS_APKA\\d shows up in both the pKa and .atoms
    return("SPARC-like")
  } else if (npKa_col == 1) {
    # ADMET Pred
    return("ADMETPred-like")
  } else {
    cat("pKa_normalization_type: Unrecognized pKa prediction format.\n")
    cat("Acceptable options {fields}:\n")
    cat("  MARVIN-like: {apKa1, ..., apKaN, pKa.atoms, bpKa1, ..., bpKaN, pKb.atoms}\n")
    cat("  ADMET Predictor-like: {S.Acidic_pKa, S.Basic_pKa, S.Mixed_pKa}\n")
    stop("Unidentifiable pKa prediction routine entered?\n")
  }
}
