##########
## Source: pKaTools.R
## Author: Cory L Strope
## Date: 1/27/2015
## Description: A set of scripts used to read essential data (fingerprints, categories) and send the data to the proper functions.
## An additional function is incorporating the code needed to remove duplicated data, and to create the pKaTools.df that is used
## for most every function (this should be in dataSet...)
##########

library(stringr)
library(gdata)
library(xlsx)
library(XLConnect)
library(ggplot2)

#####
## Take a pKaTools.df and create a full fledged data object out of it.
#####
pKaTools2ChemObject <- function(df) 
{
  data.o <- Data$new()
  data.o$pKaTools.df <- df
  ## Set chemList from df
  data.o$SetChemList(data.o$pKaTools.df)
  ## Follow up with pKa setup.
  data.o$pKaTools_Object()
  return(data.o)
}

#####
## Compare outside data source to the pKaTools.df of a data object. Save matching CASRN from pKaTools.df that are in the outside
## data source. Create new data frame with the merged data.
#####
MergeWithExternalData <- function(pKaTools.df, external.df) {
  ## Assuming the data is already read in.
  df <- data.frame()
  ## Need to deal with the CAS issue if this function proves useful enough.
  names(external.df) <- NormalizeColumnNames(names(external.df))
  if (!("CASRN" %in% names(external.df))) {
    stop("External data MUST have a case-insensitive column named {'CASRN', 'CAS', 'CASN'}.\n")
  }
  for (i in 1:nrow(pKaTools.df)) {
    incas <- searchCASRN(external.df$CASRN, pKaTools.df$CASRN[i])
    if (incas$is.true) {
      df <- rbind(df, pKaTools.df[i,])
      df$CASRN[nrow(df)] <- incas$found.cas
    }
  }

  cat("Rows in data frame:", nrow(df), "\n")
  return(df)

  out.df <- merge(df, external.df, by.x = "CASRN", by.y = "CAS")

  ## Now, output all of the merged data that correspond with the columns.
  columns.chem.data <- c("Compound", "CASRN", "Source", "SMILES", "InChI", "InChIKey", 
                       "MW", "logP", "logD", "boiling.point", "KAW", "vapor.pressure", "NOTES"
                      )
  columns.MARVIN <- c("Compound", "CASRN", "SMILES", "Major.Microspecies.at.pH.7.4", "Charge.of.MMS.at.pH.7.4",
                    "KLOP.log.P", "PHYS.log.P", "VG.log.P", "average.log.P", 
                    "KLOP.log.D.at.pH.7.4", "PHYS.log.D.at.pH.7.4", "VG.log.D.at.pH.7.4", "average.log.D.at.pH.7.4",
                    "pKaTools_APKA", "pKaTools_APKA.atoms", "pKaTools_BPKA", "pKaTools_BPKA.atoms" 
                   )
  columns.pK <- c("Compound", "CASRN", "VDss..L.kg.", "CL..mL.min.kg.", "fu", "MRT..h.", "t1.2..h.", "Reference", "Comments")

  chem.data <- out.df[,columns.chem.data]
  MARVIN <- out.df[,columns.MARVIN]
  fingerprints <- out.df[,which(names(out.df) %in% names(ia_dictionary))]
  fingerprints <- cbind(out.df$Compound, fingerprints)
  fingerprints <- cbind(out.df$CASRN, fingerprints)
  pK <- out.df[,columns.pK]

  save.xlsx("validation_obach2008.xlsx", chem.data, MARVIN, fingerprints, pK)

  return(out.df)
}

##########
## Really annoying looking for the CAS number and not knowing whether it exists in the CAS list. This will be a simple function call
## that will get a cas from a list.
##########
searchCASRN <- function(query.CASRN, CASRN.list)
{
  x <- data.frame(is.true = FALSE, found.cas = "")
  CASRNs <- unlist(strsplit(CASRN.list, ';'))
  for (j in seq_along(CASRNs)) {
    if (CASRNs[j] %in% query.CASRN) {
      x$is.true <- TRUE
      x$found.cas <- CASRNs[j]
      break
    }
  }
  return(x)
}

#####
## Write multiple objects out to a single Excel File, each object in a different worksheet.
## FROM: http://www.r-bloggers.com/quickly-export-multiple-r-objects-to-an-excel-workbook/
#####
save.xlsx <- function (file, ...)
 {
      require(xlsx, quietly = TRUE)
      objects <- list(...)
      fargs <- as.list(match.call(expand.dots = TRUE))
      objnames <- as.character(fargs)[-c(1, 2)]
      nobjects <- length(objects)
      for (i in 1:nobjects) {
          if (i == 1)
              write.xlsx(objects[[i]], file, sheetName = objnames[i])
          else write.xlsx(objects[[i]], file, sheetName = objnames[i],
              append = TRUE)
      }
      print(paste("Workbook", file, "has", nobjects, "worksheets."))
}

#####
## Determines if the input data.frame is in pKaTools format (no need to normalize).
#####
is.pKaToolsFormat <- function(pKa.df)
{
  if (all(c("pKaTools_APKA", "pKaTools_APKA.atoms", "pKaTools_BPKA", "pKaTools_BPKA.atoms") %in% names(pKa.df))) {
    return(TRUE)
  }
  return(FALSE)
}

#####
## Function to merge 2 data frames on "CASRN", and drop all columns from df2 that match with df1.
## df2 MUST be a data.table; Otherwise, with=FALSE will spawn an error.
#####
mergeAndDrop <- function(df1, df2, merge_by = "CASRN") {
  if (nrow(df1) == 0 & nrow(df2) == 0) {
    stop("I do not know how to merge 2 empty data frames. Something is wrong with your code.\n")
  }
  if (nrow(df1) == 0) {
    return(df2)
  } else if (nrow(df2) == 0) {
    return(df1)
  }

  drops <- intersect(names(df1), names(df2))
  if (length(drops) == 0) { 
    send.df <- cbind(df1, df2)
  } else {
    drops <- drops[-which(drops %in% "CASRN")]    ## Need CASRN for merge.
    df2 <- cbind(df2, DUMMY = rep(0, nrow(df2)))    ## If only CASRN is left, then this is a list, blowing up function. Keep as df using DUMMY
    x.df <- df2[, !names(df2) %in% drops, with = FALSE]
    send.df <- merge(df1, x.df)   ## Should default to CASRN (the shared key)
    send.df <- send.df[, !names(send.df) == "DUMMY"]    ## Remove the dummy column.
  }

  return(send.df)
}

setValidation <- function(validation.data, validation.data.sheet, indf = NULL) {
  valdat <- read.xls(validation.data, sheet = validation.data.sheet, header = TRUE, stringsAsFactors = FALSE)
  CAS.clean <- get.CAS.list(valdat$CASRN)
  valdat$CASRN <- CAS.clean
  valdat <- valdat[!duplicated(valdat, by = "CASRN"),]
  valdat <- valdat[order(valdat$CASRN),]
  if(!is.null(indf)) valdat <- mergeAndDrop(valdat, indf, merge_by="CASRN")
  return(valdat)
}

#####
## Set the fingerprints 
#####
setFingerprints <- function(fp_file, sheet="fingerprints", indf = NULL) {
  #skip_read <- FALSE
  #if (exists("fp")) {
    ## This is a long process, so want to make sure user has the opportunity to short circuit (if pre-calculated)
  #  user.input <- ask("Predictions for fingerprints already in memory. Proceed with current values? [Y|n] ")
  #  if (toupper(user.input) %in% c("Y", "", "YES", "YE")) { skip_read <- TRUE }
  #} 
  #if (!skip_read) {
  cat("Reading fingerprints: ", fp_file, "):\n")
  #fp <<- read.xls(fp_file, sheet = sheet, stringsAsFactors=FALSE, header=TRUE)
  fp <- read.file(fp_file, sheet=sheet, sep = "|")
  #}
  CAS.clean <- get.CAS.list(fp$CASRN)
  fp$CASRN <- CAS.clean
  fp <- fp[!duplicated(fp, by = "CASRN"),]
  setorder(fp, CASRN)
  if(!is.null(indf)) fp <- mergeAndDrop(fp, indf, merge_by="CASRN")
  return(fp)
}

## CASRN ##
get.CAS.list <- function(inCASRN) {
  CAS.list <- vector()
  for (i in 1:length(inCASRN))
    CAS.list[i] <- format_pKaTools.list(cleanCASRN(inCASRN[[i]]))
  return(CAS.list)
}

#####
## Takes a list item and returns a vector of the unique CASRNs.
#####
cleanCASRN <- function(casrn) {
  casrn <- gsub("\\n", ",", casrn)
  ## Regular expression assumes a comma-separate list. If spaces separate elements, this will not work. ##
  casrn <- unlist(strsplit(casrn, ","))

  ## This chunk of code fixes CAS #'s that Excel decides should be dates. I hate Excel for data crunching.
  if (length(casrn) == 1) {
    casparts <- unlist(strsplit(casrn, "-"))
    ## Then Excel change it into a date.
    if (length(casparts[3] > 1)) {
      x <- as.numeric(casparts[3])
      if (x > 9) stop("CAS Error: Check your Excel File for CASRN \"", casrn, "\"\nI know it is annoying, but that's Excel for you.\n")
      casrn <- paste0(casparts[1], "-", casparts[2], "-", x)
    }
  }

  casrn <- gsub("\\s+", '', casrn)   
  casrn <- gsub("^0+", '', casrn)
  return(unique(casrn))
}

canonicalizeSmiles <- function (SMILESString) {
  chemID = chemID()
  canonicalSmiles <- chemID$InChI2SMILES(chemID$SMILES2InChI(SMILESString))
  return(canonicalSmiles)
}

format_pKaTools.list <- function(delimited_list) {
  return(paste(delimited_list, collapse=";"))
}

#####
## Calculate the total number of chemicals in the input data.frame, and diagram each
#####
totalChemAndVenn <- function(in.df, fromSource = "triple-ven-chemclass") {
  ## Definition of Venn Diagram parameters:
  num.Rx <- length(which(in.df$is.Rx == TRUE))
  num.near <- length(which(in.df$is.near == TRUE))
  num.far <- length(which(in.df$is.far == TRUE))  
  Rx_near_far <-   length(which(in.df$is.Rx == TRUE  & in.df$is.near == TRUE  & in.df$is.far == TRUE))
  Rx_near_Xfar <-  length(which(in.df$is.Rx == TRUE  & in.df$is.near == TRUE  & in.df$is.far == FALSE))
  Rx_Xnear_far <-  length(which(in.df$is.Rx == TRUE  & in.df$is.near == FALSE & in.df$is.far == TRUE))
  Rx_Xnear_Xfar <- length(which(in.df$is.Rx == TRUE  & in.df$is.near == FALSE & in.df$is.far == FALSE))
  XRx_near_far <-  length(which(in.df$is.Rx == FALSE & in.df$is.near == TRUE  & in.df$is.far == TRUE))
  XRx_near_Xfar <- length(which(in.df$is.Rx == FALSE & in.df$is.near == TRUE  & in.df$is.far == FALSE))
  XRx_Xnear_far <- length(which(in.df$is.Rx == FALSE & in.df$is.near == FALSE & in.df$is.far == TRUE))
  # No False False False, since everything that is all false gets sent to far-field.
  Rx.only <- length(which(in.df$is.Rx == TRUE))
  near.only <- length(which(in.df$is.near == TRUE)) - Rx_near_Xfar - Rx_near_far
  far.only <- length(which(in.df$is.far == TRUE)) - XRx_near_far - Rx_near_far - Rx_Xnear_far
  ## non-overlapping:
  Rx.space <- c(1:Rx.only)
  near.space <- c((Rx.only+1):(Rx.only+near.only))
  far.space <- c((Rx.only+near.only+1):(Rx.only+near.only+far.only))
  ## overlapping:
  near.overlap.Rx <- c(1:Rx_near_Xfar)
  far.overlap.Rx <- c((Rx_near_Xfar+1):(Rx_near_Xfar+Rx_Xnear_far))
  near.far.Rx.overlap <- c((Rx_near_Xfar+Rx_Xnear_far+1):(Rx_near_Xfar+Rx_Xnear_far+Rx_near_far))
  far.overlap.near <- c((Rx.only+1):(Rx.only+XRx_near_far))
  near.space <- c(near.space, near.overlap.Rx, near.far.Rx.overlap)
  far.space <- c(far.space, far.overlap.Rx, near.far.Rx.overlap, far.overlap.near)

  cat("Totals:\n")
  cat("Rx:  ", Rx.only, "\n")
  cat("Near:", near.only, "\n")
  cat("Far: ", far.only, "\n")

  library(VennDiagram)
  outfile <- paste0(fromSource, ".tiff")
  venn.diagram(
    x = list(
             "Pharma"=Rx.space,
             "Near field" = near.space,
             "Far field" = far.space
            ),
      lwd=c(num.Rx,num.near,num.far),
      filename = outfile,
      resolution = 300,
      height = 3,
      width = 3,
      units = "in",
      imagetype = "tiff",
      euler.d=TRUE,
      scaled=TRUE,
      col = "transparent",
      fill = c("red","green","blue"),
      alpha=0.25,
      label.col = c("white","white","white","white","white","white","white"),
      cex=2.5,
      fontfamily="serif",
      fontface="bold",
      cat.default.pos="text",
      cat.col=c("darkred","darkgreen","darkblue"),
      cat.cex=2.5,
      cat.fontfamily="serif",
      cat.dist=c(0.06,0.06,0.03),
      cat.pos=0
    )

  cat("Printed Venn Diagram. File = \"", outfile, "\".\n")
}

DigitizeCharge <- function(charge.vector) {
  return.charge <- gsub("positive", "1", charge.vector)
  return.charge <- gsub("negative", "-1", return.charge)
  return.charge <- gsub("neutral", "0", return.charge)
  return(return.charge)
}

#####
## Replaces names in chemical data input with names that match input? Some examples:
## * replace '_' and '.' with ''
## * Replace Henrys_Law (or the like) with KAW
## * toupper()ize
#####
standardize_chemData = function(chemData_names) {
  newnames <- toupper(chemData_names)
  newnames <- lapply(newnames, FUN=function(x) { gsub("\\.", "", x)})
  newnames <- lapply(newnames, FUN=function(x) { gsub("_", "", x)})
  newnames <- unlist(newnames)
  ## Check the thesaurus for the correct fieldname.
  for (i in 1:length(newnames)) {
    newnames[i] <- LookUpSynonym(newnames[i])
  }
  ## Check for duplicate column names. If there are duplicates, then assume the first column is the correct one, and append
  ## .X to any subsequent columns, where X = {1,2,...N} for N identically named columns.
  for (i in 1:length(newnames)) {
    idx <- which(newnames[i] == newnames)
    if (length(idx) > 1) {
      suffix <- 1
      for (j in 2:length(idx)) {
        newnames[idx[j]] <- paste0(newnames[idx[j]], ".", suffix)
        suffix <- suffix + 1
      }
    }
  }

  return(newnames)
}

##########
## Set df names to PBPKxIon-used names
##########
NormalizeColumnNames <- function(names)
{
  pbpkxion.names <- vector()
  for (i in seq_along(names)) {
    pbpkxion.names[i] <- LookUpSynonym(names[i])
  }
  return(pbpkxion.names)
}

#####
## For testing the impact of a specific chemical property or feature on a particular reporter value, it is useful
## to be able to populate a data.frame with a generic chemical in chemList. This function will generate a generic
## chemical for those uses. Some things that may need to be passed in future generations will be pKa settings,
## SMILES string differences (is there an 'O' for P.vow calculations, etc.). For now, setting a neutral chemical
## with no interesting aspects, as the calculation needed right now is the impact of logP on Vdss. n is the number
## of chemicals to generate and return as a data.frame.
##
## Due to some weak correlations between parameter values and logP, we pass in the data.object and do either a 
## k-nearest neighbor or monte carlo bootstrap to generate several hypothetical chemicals for each round of the
## n total chemicals. From those chemicals, choose either the mean or median value of the set of parameters to
## use in the creation of the generic chemical.
#####
GenericChemical.by.logP = function(data.object, n.MC.draws = 100, n = 1000, acid = "", base = "") {
  ## This is straightforward
  logP <- seq(to=-21, from=26, length.out=n)    # Determined by ExpU logP min/max
  P.ow <- 10^logP
  log.P.vow <- 1.115*log10(P.ow)-1.34
  P.vow <- 10^(log.P.vow)

  generic.chem <- data.frame(logP=logP,
                             P.ow=P.ow,
                             P.vow=P.vow,
                             KAW=rep(0, n),
                             #blood.air=rep(numeric(0), n),
                             vapor.pressure=rep(0, n),
                             #Kpu.BC = rep(numeric(0), n),
                             #B.P = rep(numeric(0), n),
                             #fub = rep(numeric(0), n),
                             MW = rep(0, n),
                             pKaTools_APKA = rep(as.character(acid), n),
                             pKaTools_BPKA = rep(as.character(base), n),
                             pKaTools_APKA.atoms = rep(as.character(acid), n),
                             pKaTools_BPKA.atoms = rep(as.character(base), n),     ## Don't really care about the atoms, but functions look for it.
                             SMILES = rep("NNNNNNNNNNNNNNNNNNNNN", n),
                             stringsAsFactors=FALSE
                            )
  relevant.data <- data.object$pKaTools.df[,c("logP", "KAW", "vapor.pressure", "boiling.point", "MW", "pKaTools_APKA", "pKaTools_BPKA")]
  for (i in 1:n) {
    ## Set the chemicals to draw from. Drawing from each region of the logP range at i and i+1, except from the tail ends
    ## of the logP distribution. 0.47 is hard-coded from above.
    drawspace <- relevant.data[which(relevant.data$logP > (logP[i]-0.47) & relevant.data$logP < (logP[i]+0.47)),]
    x <- nrow(drawspace)
    if (x == 0) {
      generic.chem$KAW[i] <- 0
      generic.chem$boiling.point[i] <- 0
      generic.chem$MW[i] <- 0
      generic.chem$vapor.pressure[i] <- 0
    } else {
      ## Perform MC draws
      KAW.vec <- sample(drawspace$KAW, size = n.MC.draws, replace = TRUE)
      vapor.pressure.vec <- sample(drawspace$vapor.pressure, size = n.MC.draws, replace = TRUE)
      MW.vec <- sample(drawspace$MW, size = n.MC.draws, replace = TRUE)
      boiling.point.vec <- sample(drawspace$boiling.point, size = n.MC.draws, replace = TRUE)

      generic.chem$KAW[i] <- median(KAW.vec)
      generic.chem$vapor.pressure[i] <- median(vapor.pressure.vec)
      generic.chem$MW[i] <- median(MW.vec)
      generic.chem$boiling.point[i] <- median(boiling.point.vec)
    }
  }

  return(generic.chem)
}

##########
## Retrieval from thesaurus
##########
LookUpSynonym = function(synonym) {
  ## In the future, I would like to remove all '.' and '_'. This would make the thesaurus much smaller and manageable.
  if (!exists("thesaurus")) {
    setThesaurus()
  }
  if (synonym == "all") {
    print(thesaurus)
    return:which
  }
  SYNONYM <- toupper(synonym)
  col <- which(names(thesaurus) == SYNONYM)
  if (length(col) == 0) {
    return(synonym)
  } else {
    return(thesaurus[[col]])
  }
}

##########
## Set the thesaurus before the first instance of using it.
##########
setThesaurus = function() {
  thesaurus <<- list(
    LOGD = "logD",
    KLOPLOGDATPH74 = "logD",
    PHYSLOGDATPH74 = "logD", 
    AVERAGELOGDAT74 = "logD",
    LOGP = "logP",
    KLOPLOGP = "logP", 
    VGLOGP = "logP",
    AVERAGELOGP = "logP",
    POW = "P.ow",
    PVOW = "P.vow",
    MW = "MW", 
    MOLECULARWEIGHT = "MW", 
    KAW = "KAW", 
    HENRYSLAW = "KAW",
    WATERAIR = "KAW",
    BLOODAIR = "blood.air",
    BP = "B.P",
    KPUBC = "Kpu.BC",
    BOILINGPOINT = "boiling.point",
    VAPORPRESSURE = "vapor.pressure",
    CHARGE = "charge", 
    CHARGEOFMMSATPH74 = "charge",
    SOURCE = "Source",
    DATATYPE = "data_type",
    CHEMCLASS = "chem_class",
    CLINT = "CLint", 
    CLEARANCE = "CLint",
    FUB = "fub", 
    FUP = "fub",
    FU = "fub",
    SMILES = "SMILES",
    CANONICALSMILES = "SMILES",
    CANONICAL = "SMILES",
    INCHI = "InChI",
    INCHIKEY = "InChIKey",
    CAS = "CASRN",
    CASRN = "CASRN",
    CASN = "CASRN",
    GSID = "GSID",
    COMPOUND = "Compound",
    NAME = "Compound",
    IUPAC = "Compound.iupac",
    MOLECULENAME = "Compound.molname",
    PKATOOLSAPKA = "PKATOOLS_APKA",
    PKATOOLSAPKAATOMS = "PKATOOLS_APKA.atoms",
    PKATOOLSBPKA = "PKATOOLS_BPKA",
    PKATOOLSBPKAATOMS = "PKATOOLS_BPKA.atoms"
   )
}
