###########
## Title: utilities.R
## Description: The set of functions that contain utility functions that are necessary for pbpkxion.
## Author: Cory Strope
###########

#####
## General function for reading in both Excel and .csv/.txt files.
## INPUT: file + sheet(Excel) or sep(read.table separator)
## OUTPUT: data.table of the input file.
#####
read.file <- function(file = NULL, sheet = NULL, sep = "|")
{
	if (is.null(file)) { stop("Must provide file name to read.file\n") }
	file.extension <- getFilenameComponent(file)
	workspace <- getwd()
	if (is.excel(file.extension)) {
		if (is.null(sheet)) sheet <- 1
		txtfile <- Excel2CSV(file, sheet, sep)
	} else txtfile <- file
	## Now, either already a csv in the directory, or Excel2CSV made the csv. Read, set data.table, and done.
	DT <- read.table(file = txtfile, sep = sep, header = TRUE, stringsAsFactors = FALSE)
	setkey(setDT(DT), CASRN)
	return(DT)
}

#####
## This grabs the root filename for a file, i.e., no directories or extensions.
#####
getRootFilename <- function(file)
{
	filename <- getFilenameComponent(file, split.char = "/")	## Cuts off directory names.
	root <- getFilenameComponent(filename, component = 0)		## Removes the .xls or .txt, etc.
	return(root)
}

##########
## This function reads the file name, and returns the type, depending on the file extension
##########
getFilenameComponent <- function(file, split.char = ".", component = -1)
{
	## Get last element of the string as split by the literal '.'
	file.vector <- unlist(strsplit(file, split.char, fixed = TRUE))
	if (component == -1) component <- length(file.vector)
	else if (component == 0) component <- c(1:(length(file.vector)-1))
	return(file.vector[component])
}

is.excel <- function(file.extension)
{
	Excel.extensions <- c("xls", "xlsx")
	if (tolower(file.extension) %in% Excel.extensions) return(TRUE)
	return(FALSE)
}

#####
## Utility function that will convert excel to a CSV (by default separated by a pipe, so maybe PSV).
## This is primarily because Excel cannot easily do this function. Yet another irritation introduced by Excel.
## Writes .txt file to working directory.
#####
Excel2CSV <- function(file, sheet, sep)
{
	xl <- read.xls(file, sheet = sheet, stringsAsFactors = FALSE, header = TRUE)
	## This function counts on the fact that the extension Excel is removed already. ##
	outfile <- paste0(getwd(), "/", getRootFilename(file), "_", sheet, ".txt")
	write.table(x = xl, file = outfile, sep = sep)
	return(outfile)
}

#####
## Provided a data frame that contains CASRN, this function will take all of the chemicals, identify whether they are in
## the from.df (best case scenario, some data.object$pKaTools.df is provided, since the CASRN are curated to work within
## pKaTools). 
## OUTPUT: the pKaTools_APKA/_BPKA/_APKA.atoms/_BPKA.atoms of the chemicals that can be identified. This data is
## cbind()'ed to the query.df to make the new data frame with the pKa's.
#####
getChemList_pKa <- function(query.df, from.df)
{
	pKa.data <- data.frame()
	crap <- data.frame(SMILES = NA, pKaTools_APKA = NA, pKaTools_APKA.atoms = NA, pKaTools_BPKA = NA, pKaTools_BPKA.atoms = NA, stringsAsFactors = FALSE)
	for (i in 1:nrow(need.pka)) {
    	chemical <- getChemicalByCASRN(paste0("^", need.pka$CASRN[i]), data.object$pKaTools.df)
    	## Chemical not in from.df list
    	if (length(chemical) == 0) pKa.data <- rbind(pKa.data, crap)
    	## In list, get data
    	else pKa.data <- rbind(pKa.data, data.object$pKaTools.df[chemical, c("SMILES", "pKaTools_APKA", "pKaTools_APKA.atoms", "pKaTools_BPKA", "pKaTools_BPKA.atoms")])
	}
	out.df <- cbind(query.df, pKa.data)
	return(out.df)
}

getChemicalByCASRN <- function(query, from.df)
{
	found <- vector()
	for (i in 1:nrow(from.df)) {
		row.cas <- unlist(strsplit(from.df$CASRN[i], ";"))
		## If any of the synonyms are true, then we "found" the chemical.
		if ( any(grepl(pattern = paste0("^", query), x = row.cas)) ) found <- c(found, i)
	}
	return(found)
}