####################
##### Chemical properties class. Holds all necessary information for a chemical.
##### * Question marks:
#####   - What about metabolism (future)
####################
chemicalProperties <- setRefClass("chemicalProperties",
  fields = list(
    ID = "chemID",
    structure = "chemStructure",
    microspecies = "list",
    pKa = "pKa",
    logD = "numeric",
    logP = "numeric",
    P.ow = "numeric",
    P.vow = "numeric",
    MW = "numeric",
    KAW = "numeric",
    blood.air = "numeric",
    B.P = "numeric",
    boiling.point = "numeric",
    vapor.pressure = "numeric",
    charge = "numeric",
    Source = "character",
    data_type = "character",
    chem_class = "character",
    cpcat_term = "character",
    CLint = "numeric",
    fub = "numeric",
    is.Rx = "logical",
    is.near = "logical",
    is.far = "logical",
    Kpu.BC = "numeric",
    is.HLOC = "logical",    ## decoupling of logP and HLOC-testing; helpful for turning off HLOC/Peyret PC 
    is.Peyret = "logical",  ## above
    thesaurus = "list"     ## A list of many different "synonyms" of chemical property fields. ##
  ),
  methods = list(
     #####
    ## General function that can set all character/numeric data types based on an incoming data.frame with colnames
    ## that match the chemicalProperty class fields names above.
    #####
    populateChemicalPropertyFields = function(chemRow=data.frame) {
      ### Segregate chemical identifier columns to new df (pcp.df)
      # Get the column names referring to identifiers (object chemID fillers)
      chemIdentifiers <- toupper(colnames(chemRow)) %in% toupper(ID$getIdentifierColnames())
      if (sum(chemIdentifiers, na.rm=TRUE)) {
        ## --with, from data.table package--
        ## with: By default 'with=TRUE' and 'j' is evaluated within the frame of 'x'. 
        ## The column names can be used as variables. When 'with=FALSE', 'j' works as it does in '[.data.frame'.
        ID_df <- chemRow[,chemIdentifiers, with=FALSE]
        #print(ID_df)
        .self$ID$setIdentifiers(ID_df)
        #cat("ID data_frame!: \n") 
        chemRow <- chemRow[,!chemIdentifiers, with=FALSE]
      }

      chemIdentifiers <- toupper(colnames(chemRow)) %in% toupper(structure$getIdentifierColnames())
      if (sum(chemIdentifiers, na.rm=TRUE)) {
        struct_df <- chemRow[,chemIdentifiers, with=FALSE]
        .self$structure$setIdentifiers(struct_df)
        chemRow <- chemRow[,!chemIdentifiers, with=FALSE]
      }

      ## Upon read-in, the '.' is removed in variables that have them. Here, we compare the dot-free input variables to the dotted
      ## class fields.
      ## vals2set is a data.frame of 1 row. To access the actual data (rather than the row), need to specify the ith element of vals2set.
      chemRow <- chemRow[,which(names(chemRow) %in% names(.self$.refClassDef@fieldClasses)), with = FALSE]
      if (length(chemRow) > 0) {
        for (i in 1:length(chemRow)) {
          ## --------------------------------
          ## Below comments defunct as of 20150224: Created thesaurus that set all synonymous fields to the chemicalProperties
          ## field name.
          ## --------------------------------
          ## Breakdown of following code:
          ## Each input value from the chem.data file is set to a particular class(), which may not agree with the slot value
          ## in the refClass. In order to fix this, the code (1) finds the appropriate class slot where the input value is located, e.g.,
          ## Source is slot #13 (on 20150121), (2) set the slot value to the input value. 
          ## For step 2, the code makes sure that the class() of the slot and the input value are the same. This is done on the fly by
          ## setting the right side of the '<-' using the R casting system of as.XXXXXX, where XXXXXX is found by looking at the class
          ## of the value in the slot whose name matches column name of vals2set. 
          ## --------------------------------
          if (!is.na(chemRow[1,i, with = FALSE])) {
            to_eval <- paste0(".self$", names(chemRow)[i], "<- as.", .self$.refClassDef@fieldClasses[names(chemRow)[i]], "(chemRow[1,", i, ", with = FALSE])")
            eval(parse(text=to_eval))
          }
        }
      }
      .self$FillHydroAffinity()
      ## Should now have all the data for Buist2012 P.ba
      .self$blood.air <- Buist2012_Pba(.self)
    },
    #####
    ## Populate P.ow, P.vow, and logP based on input.
    #####
    FillHydroAffinity = function() {
      set.param.logP <- FALSE
      set.param.Pow <- FALSE
      set.param.Pvow <- FALSE
      set.param.logP <- if (length(.self$logP) == 1) FALSE else TRUE
      set.param.Pow <- if (length(.self$P.ow) == 1) FALSE else TRUE
      set.param.Pvow <- if (length(.self$P.vow) == 1) FALSE else TRUE
      ## If you have to set all params: Little can be done if we do not have the water PC.
      if (set.param.logP & set.param.Pow & set.param.Pvow) {
        stop("PBPKxIon Error:: No water affinity variable found (logP, P.ow, P.vow).\n")
      }
      if (!set.param.logP) {    ## logP is set (i.e., do not have to set param logP)
        if (set.param.Pow) { 
          .self$P.ow <- 10^(.self$logP)
        }
        if (set.param.Pvow) {
          .self$P.vow <- Pvow(.self$P.ow, SMILES = .self$structure$SMILES)
        }
      } else if (!set.param.Pow) {
        .self$logP <- log10(.self$P.ow)
        if (set.param.Pvow) {
          .self$P.vow <- Pvow(.self$P.ow, SMILES = .self$structure$SMILES)
        }
      } else if (!set.param.Pvow) {
        ## Never heard of the vow begin set without any of the others, but for those that are, assume Pvow=Pow=10^logP
        .self$P.ow <- .self$P.vow
        .self$logP <- log10(.self$P.vow)
      }
      if (.self$logP > 5.8) {
        .self$is.HLOC <- TRUE
        .self$is.Peyret <- FALSE
      } else {
        .self$is.Peyret <- TRUE
        .self$is.HLOC <- FALSE
      }
    },
    #####
    ## Utilize the ToxCast and Tox21 vectors in the RData files to set source to ToxCast, Tox21, or nothing.
    #####
    setSource = function() {
      myCAS <- .self$ID$vectorize.CAS()
      if(exists("ToxCast")) {
        inToxCast <- which(myCAS %in% ToxCast)
        if (length(inToxCast) > 0) {
          .self$Source <- "ToxCast"
        } else {
          inTox21 <- which(myCAS %in% Tox21)
          if (length(inTox21) > 0) {
            .self$Source <- "Tox21"
          } else .self$Source <- "user_data"
        } 
      } else .self$Source <- "user_data"
    },
    getChemPKA = function(which.pKa) {
      ## Make a data.frame to grab all of the individual pKa attributes. ##
      df <- data.frame(value=pKa$getPKAValue(which.pKa),
                       SMILES=pKa$getPKASMILES(which.pKa),
                       SMILES_position=pKa$getPKASMILESPosition(which.pKa),
                       class=pKa$getPKAClass(which.pKa),
                       stringsAsFactors=FALSE
                      )
      #cat("getChempKa input, which.pKa:", which.pKa, "\n")
      return(df)
    },
    #####
    # This populates the pKa data.frame for each chemical.
    #####
    populatePKA = function(chemRow=data.frame) {
      ## If chemical has no pKa, that is fine. Just have a pKa data frame without any chemical.
      ## Incoming data.frame will have 4 columns of interest: 
      # (1,2) pKaTools_APKA, pKaTools_APKA.atoms
      # (3,4) pKaTools_BPKA, pKaTools_BPKA.atoms
      # These are formatted as semicolon-separated lists. Isolate 1,2,3,4, pass to setpKa...?
      #print(chemRow)
      #cat("#################pka#################\n")
      #cat("## populatePKA\n")
      pKa2pass <- chemRow[,c("pKaTools_APKA", "pKaTools_BPKA", "pKaTools_APKA.atoms", "pKaTools_BPKA.atoms")]
      #cat("## ")
      #print(pKa2pass)
      #cat("## SMILES:", .self$structure$SMILES,"vs structure$SMILES:", structure$SMILES, "vs chemRow$SMILES:", chemRow$SMILES, "\n")
      #cat("#################pka#################\n")
      .self$pKa$setpKa(pKa2pass, chemRow$SMILES)
    },
    #####
    ## Return the number of pKas/ionizable.atoms for this chemical 
    #####
    getNumIonizableSites = function() {
      return(nrow(pKa$pKa_df))
    },
    #####
    ## Get the fraction of species that is neutral
    ## - Right now ignoring that strong zwitterions can be "neutral" in 2 ways: no charged atoms || pos atoms = neg atoms
    #####
    FractionIonizedList = function(pH = 7.4) {
      frac <- list(negative = 0, neutral = 0, positive = 0)
      apKa <- NULL
      bpKa <- NULL
      base.length <- 0
      i__pKa__ <- NULL    ## If "", gets turned into NA.
      if (any(.self$pKa$pKa_df$class == ACID)) {
        apKa <- sort(.self$pKa$pKa_df$value[which(.self$pKa$pKa_df$class == ACID)])
        i__pKa__ <- c(i__pKa__, apKa)
        apKa <- paste(apKa, collapse=";")
      }
      if (any(.self$pKa$pKa_df$class == BASE)) {
        bpKa <- sort(.self$pKa$pKa_df$value[which(.self$pKa$pKa_df$class == BASE)], decreasing=TRUE)
        base.length <- length(bpKa)   ## For strong zwit
        i__pKa__ <- c(i__pKa__, bpKa)
        bpKa <- paste(bpKa, collapse=";")
      }

      which.species <- getFracSpecNumerator(apKa, bpKa)
      if (which.species == -1) {
        ## Neutrals.
        frac$negative <- 0
        frac$positive <- 0 
        frac$neutral <- 1
      } else {
        #####
        ## Charged species: i__pKa__ != NULL (which causes error upon sorting)
        #####
        ## First, strong zwitterions need different species assigned.
        if (which.species == -2) {
        which.species <- base.length
        }
        ## List needs to be sorted for the frac_species call which uses Henderson Hasselbalch.
        i__pKa__ <- as.numeric(sort(i__pKa__))
        frac$neutral <- .self$pKa$frac_species(pH, which.species, i__pKa__)
        if (which.species == 0) {
          ## Acids
          frac$positive <- 0
          frac$negative <- 1 - frac$neutral   ## Because acidic, so this is real.
        } else {
          ## This category holds bases, weak zwitterions, and strong zwitterions.
          ## Again, this is iffy. Strong zwitterions can have both positive and negative charges
          frac$negative <- 0
          frac$positive <- 1 - frac$neutral
        }
      }
      return(frac)
    },
    #####
    ## Calculate the ionization constant per Henderson-Hasselbalch for a chemical
    #####
    I.m = function(pH = 7.4) {
      ## For the given chemical type & pKas, calculate the Henderson.Hasselbalch constants
      ## Chemical type is irrelevant. If there is a distant base and a close acid, the acid will overpower any contributions of basic
      ## pKas
      I.m <- Henderson.Hasselbalch(pH)
      return(I.m)
    },
    #####
    ## Calculation of the distribution constant
    #####
    f.logD = function(pH) {
      .self$logD <- .self$logP - log10(1 + Henderson.Hasselbalch(pH))
    },
    #####
    ## Calculating the ionization constant for chemicals that have acceptable parameters.
    #####
    Henderson.Hasselbalch = function(pH) {
      ## Sort out the acidic and basic pKa's of the chemical
      if (any(pKa$pKa_df$class == ACID)) {
        acid.pKa <- sort(pKa$pKa_df$value[which(pKa$pKa_df$class == ACID)])
        omnibus.acid <- recursive.HH(0, 0, acid.pKa, pH, ACID)
      } else omnibus.acid <- 0
      if (any(pKa$pKa_df$class == BASE)) {
        basic.pKa <- sort(pKa$pKa_df$value[which(pKa$pKa_df$class == BASE)], decreasing=TRUE)
        omnibus.base <- recursive.HH(0, 0, basic.pKa, pH, BASE)
      } else omnibus.base <- 0
      ## Acid then base: Recursive call, where variables are: Current value, current pKa, pKa vector, pH, and acid ("A") or base ("B")
      ## If straight acid or base, then base=0 or acid=0, respectively. If zwitterion, then this calculation is still correct.
      omnibus.val <- omnibus.acid + omnibus.base
      #cat("    Omnibus.val returned from pKa values (", pKa$value, "; acid=", acid.pKa, "; base=", basic.pKa, "): ", omnibus.val, "\n")
      return(omnibus.val)
    },
    #####
    ## Recursive function for calculating the Henderson-Hasselbalch equation for any number of pKa's
    #####
    recursive.HH = function(
                            prev.val,      ## Previous value from the recursion
                            depth,         ## Add one each time (times through the recursive.HH function
                            pKa,           ## pKa values
                            pH,            ## pH under consideration
                            general.class  ## ACID or BASE
                           ) 
    {
      ## Stopping condition.
      if (depth == length(pKa)) return(0)

      ## Otherwise, continue calculation.
      depth <- depth + 1
      if (general.class == ACID) new.term <- pH - pKa[depth]     ## Acid
      else new.term <- pKa[depth] - pH                           ## Base
      new.val <- prev.val + new.term
      return(10^(new.val) + recursive.HH(new.val, depth, pKa, pH, general.class))
    }
  )
)


