##########
### Class definition for chemical SMILES strings and associated functions.
##########
SMILES <- setRefClass("SMILES",
  fields = list(
    SMILES="vector"
  ),
  methods = list(
    #####
    ## Several SMILES strings have constantly ionized sites (e.g., 50-10-2 and 50-65-7). This will cause problems 
    ## identifying the site that ionization acts upon for the ionizedSMILES in the pKa_df for chemicals. This routine 
    ## removes all such ionizations from the SMILES strings.
    #####
    clearIonization = function(work_SMILES) {
      ## Or together all of the possible items to remove.
      ionization_symbols <- c("[", "]", "+", "-", "H")
      to_zap <- which(work_SMILES %in% ionization_symbols)

      if(length(to_zap) > 0)
        work_SMILES <- work_SMILES[-to_zap]

      return(work_SMILES)
    },
    #####
    ## Accessory function that performs strsplit(x,'') for compounds. This re-write splits by atom (or SMILES element)
    #####
    compoundStrSplit = function(inSMILES = .self$SMILES) {
#      tmp <- unlist(strsplit(.self$SMILES, ''))
      tmp <- unlist(strsplit(inSMILES, ''))
      work <- NULL
      #cat(SMILES, "  ", length(tmp), "  ", tmp, "\n")
      if (length(tmp) > 1) {
        for (i in 1:(length(tmp)-1)) {
          if (!is.na(tmp[i])) {
            if (tmp[i] != "r" & tmp[i] != "l" & tmp[i] != "i") {
              atom <- tmp[i]
              #cat(i, ": ", atom, "  i+1: ", tmp[i+1], "\n")
              if(!is.na(tmp[i+1])) {
                if (tmp[i] == "C" | tmp[i] == "c") if (tmp[i+1] == "l") atom <- paste0(tmp[i], "l") 
                if (tmp[i] == "B" | tmp[i] == "b") if (tmp[i+1] == "r") atom <- paste0(tmp[i], "r")
                if (tmp[i] == "S" | tmp[i] == "s") if (tmp[i+1] == "i") atom <- paste0(tmp[i], "i")
                if (tmp[i] == "S" | tmp[i] == "s") if (tmp[i+1] == "e") atom <- paste0(tmp[i], "e")
                if (tmp[i] == "A" | tmp[i] == "a") if (tmp[i+1] == "s") atom <- paste0(tmp[i], "s")
              }
              work <- c(work, atom)
            }
          }
        }
      }
      ## Now, add last atom if it is not an 'r' (Bromine suffix) or an 'l' (Chlorine suffix) or 'i' (Silicon suffix)
      if (tmp[length(tmp)] != "r" & tmp[length(tmp)] != "l" & tmp[length(tmp)] != "i") work <- c(work, tmp[length(tmp)])

      return(work)
    }
  )
)

SMILES$methods(
  #####
  ## Getting elements from SMILES strings.
  #####
  splitSMILES = function() {
    work <- compoundStrSplit()
    work <- clearIonization(work)
    return(work)
  },
  #####
  ## Returns the number of non-hydrogen atoms in the compound.
  #####
  atomCount = function() {
    work <- compoundStrSplit()
    work <- clearIonization(work)
    atom.idx <- findAtoms(work)  ## Locate atom characters, not structure characters
    return(length(atom.idx))
  },
  #####
  ## Singles out an atom for finding if there are hydrogens
  #####
  hydrogenCount = function(atom.num) {
    ## Atom is the specific atom of interest. Prefix and postfix are the previous and subsequent sequence data
    smiles.df <- data.frame(prefix = "", atom = "", suffix = "", smiles.position = 0)
    smiles.elements <- splitSMILES()
    atom.idx <- 0
    pre <- TRUE
    found <- FALSE
    smiles.position <- 0
    for (i in 1:length(smiles.elements)) {
      if (smiles.elements[i] %in% chemical_atoms) atom.idx <- atom.idx + 1
      if (atom.idx == atom.num & !found) {
        smiles.df$atom <- smiles.elements[i]
        smiles.df$smiles.position <- i
        ## If atom is followed by non-atom SMILES character, then want to add to suffix.
        ## First pass (atom found), this is not active. subsequent passes, all non-atom elements
        ## will be added.
        if (found) {
          smiles.df$suffix <- paste0(smiles.df$suffix, smiles.elements[i])
        }
        found <- TRUE
        pre <- FALSE
      } else if (pre) {
        smiles.df$prefix <- paste0(smiles.df$prefix, smiles.elements[i])
      } else {
        smiles.df$suffix <- paste0(smiles.df$suffix, smiles.elements[i])
      }
    }

    #####
    ## Each element now identified, so now to determine if there are hydrogens attached.
    #####
    atoms <- c(6, 6, 2, 2, 3, 3, 4, 4, 4, 4)
    names(atoms) <- c("S", "s", "O", "o", "N", "n", "C", "c", "P", "p")
    lower.atoms <- c("s", "c", "n", "o", "p")
    upper.atoms <- c("S", "C", "N", "O", "P")
    ring.nums <- c("1", "2", "3", "4", "5", "6", "7", "8", "9")
    num.covalent <- as.integer(atoms[which(names(atoms) == smiles.df$atom)])

    ## Have the number of covalent bonds available, now need to count the bonds already attached.
    ## If, after subtracting num.covalent, there are still available bonds, these are hydrogens.
    # pre
    prefix.contrib <- 0
    pre.seq <- 0
    pre.seq1 <- ""
    if (smiles.df$prefix == "") {
      prefix.contrib <- 0
    } else {
      pre.seq <- compoundStrSplit(smiles.df$prefix)
      pre.seq1 <- pre.seq[length(pre.seq)]
      #cat("pre.seq1", pre.seq1, "\n"  )
      if (pre.seq1 %in% names(atoms)) {
        ## It is an atom, prefix contributes 1
        prefix.contrib <- 1
      } else if (pre.seq1 == "=") {
        prefix.contrib <- 2
      } else if (pre.seq1 == "(") {
        prefix.contrib <- 1
      } else if (pre.seq1 == ")") {
        prefix.contrib <- 1
      } else if (pre.seq1 %in% ring.nums) {
        prefix.contrib <- 1
      }
    }

    ###       12 3 4567 8  9
    ### CASE: CC(O)CCCC(C)=C
    ### atom 7 has 4 bonds, but unless read to "=", looks only like 3. I do need to make the read-forward routine.

    smiles.position.of.atom <- length(pre.seq) + 1 ## <prev.seq> + <curr.atom> = SMILES position.
    #cat(smiles.position.of.atom, "  --->>> ", smiles.df$smiles.position, "\n")
    suffix.contrib <- 0
    if (smiles.df$suffix == "") suffix.contrib <- 0
    else {
      suf.seq <- compoundStrSplit(smiles.df$suffix)
      suf.len <- length(suf.seq)
      suf.seq1 <- suf.seq[1]
      ## Need a look-ahead for the suffix sequence.
      if (suf.len > 1) suf.seq2 <- suf.seq[2]
      else suf.seq2 <- ""

      if (suf.seq1 == "(") {
        suffix.contrib <- 2 + extraBonds(smiles.position.of.atom)
      } else if (suf.seq1 == ")") {
        suffix.contrib <- 0
      } else if (suf.seq1 %in% ring.nums) {
        suffix.contrib <- 2
        ## unless ....
        if (suf.seq2 == "(") suffix.contrib <- 3    ## Ring + second bond
      } else if (suf.seq1 == "=") {
        suffix.contrib <- 2
      } else {
        suffix.contrib <- 1
      }
    }

    if(smiles.df$atom %in% lower.atoms) {
      if (pre.seq1 %in% upper.atoms) {
        occupied <- 4
      } else {
        occupied <- 3
      }
      ### Need to do some quick checking for pre and post seq:
      if (suffix.contrib > 2) occupied <- 4
    } else {
      occupied <- (prefix.contrib + suffix.contrib)
    }

    num.covalent <- as.integer(atoms[which(names(atoms) == smiles.df$atom)])
    num.hydrogens <- num.covalent - occupied
    #cat("Atom: ", smiles.df$atom, "    available.H = ", num.covalent, " - (pre(", prefix.contrib, ") + suf(", suffix.contrib, ")) = ", num.hydrogens, "\n")

    if (length(num.hydrogens) == 0) {
      num.hydrogens <- 0
      cat("num.hydrogens is empty, atom: ", atom.num, "of SMILES ", .self$SMILES)
    }

    return(num.hydrogens)
  },
  #####
  ## branch points immediately incur a price of 2 for the branch. This read-ahead adds additional charges
  ## if there is a double bond at the end of the branch  point.
  #####
  extraBonds = function(atom.num) {
    smiles.elements <- splitSMILES()
    next.element <- checkBranch(smiles.elements, atom.num)
    ## Now determine how much the next element adds to the bond count.
    if (length(next.element) == 0) { return(0) }
    else if (smiles.elements[next.element] == '=') { return(1) }
    else if (smiles.elements[next.element] == '(') { return(0) }
    else if (smiles.elements[next.element] == ')') { return(0) }
    else { return(0) }  ## An element
  },
  checkBranch = function(inSMILES, pos) {
    #cat("checkBranch", inSMILES, pos, " // ", inSMILES[pos], " <+1>", inSMILES[pos+1], "\n")
    if (inSMILES[pos+1] != '(')
      return(pos)
    x <- Read2Paren(inSMILES, pos+2, 0)
    #cat("Returning: ", x, "\n")
    return(x)
  },
  #####
  #####
  Read2Paren = function(inSMILES, pos, recurse.depth) {
    #cat("Read2Paren", inSMILES, "   pos: ", pos, "   depth:", recurse.depth, " --> ", inSMILES[pos], "\n")
    do.more <- TRUE
    i <- pos + 1
    ## No need to read to closing paren. ##
    ## Otherwise, start reading. ##
    while(do.more) {
      #cat("position:", i, "SMILES.char:", inSMILES[i], "\n")
      if (inSMILES[i] == '(') {
        # Recurse
        #cat("    RECURSE\n")
        i <- Read2Paren(inSMILES, i, recurse.depth + 1)
        next
      } else if (inSMILES[i] == ')') {
        #cat("    RETURN i: ", i, "\n")
        #if (recurse.depth < 2) return(i)
        #else return(i+1)
        return(i+1)
      } else if (i == length(inSMILES)) {
        stop("Invalid SMILES code; no closing ')' for branch point.\n")
      } else if (recurse.depth == 0) {
        return(i+1)
      }else {
        #cat("    NEXT...", i+1, "\n")
        i <- i+1
      }
    }
  },
  #####
  ## This function ionizes a particular atom based on my knowledge of how ionizeds sites are represented in SMILES.
  ## It deconstructs the SMILES string and chooses the ionizable atom as specified by the input pKa row.
  ## OUTPUT: SMILES string with the charge as specified by the pKa information.
  #####
  ionize = function(at.pos, ion.type) {
    work <- compoundStrSplit()
    if (any(grepl("[", work, fixed=TRUE))) work <- clearIonization(work)
    # H in patterns NOT counted.
    ## Atoms with 2 characters need to be re-merged.
    atom.idx <- findAtoms(work)  ## Locate atom characters, not structure characters
    atom <- work[atom.idx[at.pos]]
    context_left <- work[atom.idx[at.pos]-1]
    context_right <- work[atom.idx[at.pos]+1]
    type <- ion.type

    #cat(atom, context_left, context_right, type, "\n")
    #cat("WORK STRING:")
    #print(work)

    ### BELOW WE NEED TO PUT THE STRING BACK TOGETHER WITH THE IONIZATION AND RETURN IT.
    work[atom.idx[at.pos]] <- getIonizedAtom(atom, context_left, context_right, type)
    ionized_smiles <- paste(work, collapse='')
    #cat("Ionized: ", ionized_smiles, "\n")
    return(ionized_smiles)
  },
  #####
  ## This returns all of the positions containing atoms from a splitted SMILES string vector.
  #####
  findAtoms = function(splittedSMILES) {
    atom.idx <- which(splittedSMILES %in% chemical_atoms)
    return(atom.idx)
  },
  #####
  ## This routine determines the ionizable atoms, and outputs the appropriate charge number of Hydrogens on the atom
  ## based on the atom type and its context in the compound.
  #####
  getIonizedAtom = function(atom, context_left, context_right, type) {
    #cat("atom:", atom, "\ncontext_left:", context_left, "\ncontext_right:", context_right, "\ntype:", type, "\n")
    chemSingleBond <- FALSE
    chemDoubleBond <- FALSE
    ## First, take care of SMILES begin/end cases (set to no value)
    if (length(context_left) == 0L)  
      context_left <- ""      ## Special case: ionizable atom at front of SMILES, character(0) ##
    if (is.na(context_right) || is.null(context_right)) context_right <- ""
    if (is.na(context_left) || is.null(context_left)) context_left <- ""
    ####### Calculate the number of bonds for each atom, output structure based on num_bonds. ######
    ## Branching points exist on left
    if (context_left == ")" || context_left == "(") { branch_left <- TRUE }
    else { branch_left <- FALSE }
    ## Right
    if (context_right == "(" || context_right == ")") { branch_right <- TRUE }
    else { branch_right <- FALSE }
    ## Explicit double-bonds: Aliphatic sections.
    if (context_left == "=") { doublebond_left <- TRUE }  # Exist
    else { doublebond_left <- FALSE }                     # Or not.
    if (context_right == "=") { doublebond_right <- TRUE } 
    else { doublebond_right <- FALSE }
    if (context_left == "#") { hash_left <- TRUE } 
    else { hash_left <- FALSE }
    if (context_right == "#") { hash_right <- TRUE } 
    else { hash_right <- FALSE }
    ## Implicit double-bonds: Aromatic rings have 3 bonds for each atom in the ring.
    if (context_left == tolower(context_left)) { aromatic_left <- TRUE }  ## Aromatic rings in SMILES are lowercase.
    else { aromatic_left <- FALSE }
    if (context_right == tolower(context_right)) { aromatic_right <- TRUE } 
    else { aromatic_right <- FALSE }
    ## Within multi-ring structure  (both left and right are ring *numbers*; see Alprazolam)
    if (!is.na(suppressWarnings(as.numeric(context_left))) && !is.na(suppressWarnings(as.numeric(context_right)))) { 
      aromatic_left <- TRUE
    }
    if (context_left == "" & (context_right != "=" | context_right != "#")) chemSingleBond <- TRUE
    if (context_left == "" & (context_right == "=" | context_right == "#")) chemDoubleBond <- TRUE
    if (context_right == "" & (context_left != "=" | context_left != "#")) chemSingleBond <- TRUE
    if (context_right == "" & (context_left == "=" | context_left == "#")) chemDoubleBond <- TRUE
    if (type == BASE) {
    if (atom == "N") {
      ## Now need to deal with the branching scheme in SMILES... The more branches, the fewer the hydrogens:
      if (chemSingleBond) return("[NH3+]")
      if (chemDoubleBond) return("[NH2+]")
        if ( (branch_left && !branch_right) || ( !branch_left && branch_right) 
              || aromatic_left || aromatic_right
           ) return("[NH+]")  ## Branch on one side or within an aromatic ring. ##
        else if (branch_left && branch_right 
                 || doublebond_left || doublebond_right
                 || hash_left
                )  return("[N+]") ## Branch on both sides ##
        else return("[NH2+]")   ## aliphatic ##
      }
      ## Also want to take care of case where N is explicitly represented in an aromatic ring. ##
      if (atom == "n") return("[nh+]") 
      if (atom == "O") {
        if (branch_right && doublebond_left) return("[O+]")
        if (branch_left) return("[OH+]")
        else return("[O+]")
      }
    } else if (type == ACID) {
      #cat(" atom =", atom, " left =", context_left, " right =", context_right, "\n")
      if (atom == "N") return("[N-]")
      if (atom == "O") return("[O-]")
      if (atom == "S") return("[S-]")
      if (atom == "B") return("[B-]")   ## Boron ##
      if (atom == "C") return("[C-]")   ## See Cefazolin ##
    }
    ## If this point is reached, then something funny happened, so we punt and just return the unionized atom.
    return(atom)  
  }
)